from copy import deepcopy
from abc import ABC, abstractmethod
from typing import List, Union, Tuple
from collections import OrderedDict

import pandas as pd

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import Select, Identifier, Star, BinaryOperation, Constant, Join, Function
from mindsdb_sql_parser.utils import JoinType
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.utilities.functions import resolve_table_identifier
from mindsdb.api.executor.utilities.sql import get_query_tables
from mindsdb.utilities.exception import EntityNotExistsError
import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.context import context as ctx
from mindsdb.api.executor.datahub.classes.tables_row import (
    TABLES_ROW_TYPE,
    TablesRow,
)


class LogTable(ABC):
    """Base class for 'table' entitie in internal 'log' database

    Attributes:
        name (str): name of the table
        deletable (bool): is it possible to delete a table
        visible (bool): should be table visible in GUI sidebar
        kind (str): type of the table/view
    """

    name: str
    deletable: bool = False
    visible: bool = True
    kind: str = 'table'

    @staticmethod
    @abstractmethod
    def _get_base_subquery() -> Select:
        """Get a query that returns the table from internal db

        Returns:
            Select: 'select' query that returns table
        """
        pass

    @staticmethod
    def company_id_comparison(table_a: str, table_b: str) -> BinaryOperation:
        """Make statement for 'safe' comparison of company_id of two tables

        Args:
            table_a (str): name of first table
            table_b (str): name of second table

        Returns:
            BinaryOperation: statement that can be used for 'safe' comparison
        """
        return BinaryOperation(
            op='=',
            args=(
                Function(
                    op='coalesce',
                    args=(
                        Identifier(f'{table_a}.company_id'),
                        0
                    )
                ),
                Function(
                    op='coalesce',
                    args=(
                        Identifier(f'{table_b}.company_id'),
                        0
                    )
                )
            )
        )


class LLMLogTable(LogTable):
    name = 'llm_log'

    columns = [
        'API_KEY', 'MODEL_NAME', 'INPUT', 'OUTPUT', 'START_TIME', 'END_TIME',
        'PROMPT_TOKENS', 'COMPLETION_TOKENS', 'TOTAL_TOKENS', 'SUCCESS'
    ]

    types_map = {
        'SUCCESS': 'boolean',
        'START_TIME': 'datetime64[ns]',
        'END_TIME': 'datetime64[ns]'
    }

    @staticmethod
    def _get_base_subquery() -> Select:
        query = Select(
            targets=[
                Identifier('llm_log.api_key', alias=Identifier('api_key')),
                Identifier('predictor.name', alias=Identifier('model_name')),
                Identifier('llm_log.input', alias=Identifier('input')),
                Identifier('llm_log.output', alias=Identifier('output')),
                Identifier('llm_log.start_time', alias=Identifier('start_time')),
                Identifier('llm_log.end_time', alias=Identifier('end_time')),
                Identifier('llm_log.prompt_tokens', alias=Identifier('prompt_tokens')),
                Identifier('llm_log.completion_tokens', alias=Identifier('completion_tokens')),
                Identifier('llm_log.total_tokens', alias=Identifier('total_tokens')),
                Identifier('llm_log.success', alias=Identifier('success'))
            ],
            from_table=Join(
                left=Identifier('llm_log'),
                right=Identifier('predictor'),
                join_type=JoinType.LEFT_JOIN,
                condition=BinaryOperation(
                    op='and',
                    args=(
                        LLMLogTable.company_id_comparison('llm_log', 'predictor'),
                        BinaryOperation(
                            op='=',
                            args=(
                                Identifier('llm_log.model_id'),
                                Identifier('predictor.id')
                            )
                        )
                    )
                )
            ),
            where=BinaryOperation(
                op='is' if ctx.company_id is None else '=',
                args=(Identifier('llm_log.company_id'), Constant(ctx.company_id))
            ),
            alias=Identifier('llm_log')
        )
        return query


class JobsHistoryTable(LogTable):
    name = 'jobs_history'

    columns = ['NAME', 'PROJECT', 'RUN_START', 'RUN_END', 'ERROR', 'QUERY']
    types_map = {
        'RUN_START': 'datetime64[ns]',
        'RUN_END': 'datetime64[ns]'
    }

    @staticmethod
    def _get_base_subquery() -> Select:
        query = Select(
            targets=[
                Identifier('jobs.name', alias=Identifier('name')),
                Identifier('project.name', alias=Identifier('project')),
                Identifier('jobs_history.start_at', alias=Identifier('run_start')),
                Identifier('jobs_history.end_at', alias=Identifier('run_end')),
                Identifier('jobs_history.error', alias=Identifier('error')),
                Identifier('jobs_history.query_str', alias=Identifier('query'))
            ],
            from_table=Join(
                left=Join(
                    left=Identifier('jobs_history'),
                    right=Identifier('jobs'),
                    join_type=JoinType.LEFT_JOIN,
                    condition=BinaryOperation(
                        op='and',
                        args=(
                            LLMLogTable.company_id_comparison('jobs_history', 'jobs'),
                            BinaryOperation(
                                op='=',
                                args=(
                                    Identifier('jobs_history.job_id'),
                                    Identifier('jobs.id')
                                )
                            )
                        )
                    )
                ),
                right=Identifier('project'),
                join_type=JoinType.LEFT_JOIN,
                condition=BinaryOperation(
                    op='and',
                    args=(
                        LLMLogTable.company_id_comparison('project', 'jobs'),
                        BinaryOperation(
                            op='=',
                            args=(
                                Identifier('project.id'),
                                Identifier('jobs.project_id')
                            )
                        )
                    )
                )
            ),
            where=BinaryOperation(
                op='is' if ctx.company_id is None else '=',
                args=(Identifier('jobs_history.company_id'), Constant(ctx.company_id))
            ),
            alias=Identifier('jobs_history')
        )
        return query


class LogDBController:
    def __init__(self):
        self._tables = OrderedDict()
        self._tables['llm_log'] = LLMLogTable
        self._tables['jobs_history'] = JobsHistoryTable

    def get_list(self) -> List[LogTable]:
        return list(self._tables.values())

    def get(self, name: str = None) -> LogTable:
        try:
            return self._tables[name]
        except KeyError:
            raise EntityNotExistsError(f'Table log.{name} does not exists')

    def get_tables(self) -> OrderedDict:
        return self._tables

    def get_tables_rows(self) -> List[TablesRow]:
        return [
            TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=table_name)
            for table_name in self._tables.keys()
        ]

    def query(self, query: Select = None, native_query: str = None,
              session=None, return_as: str = 'split') -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
        if native_query is not None:
            if query is not None:
                raise Exception("'query' and 'native_query' arguments can not be used together")
            query = parse_sql(native_query)
        else:
            query = deepcopy(query)

        if type(query) is not Select:
            raise Exception("Only 'SELECT' is allowed for tables in log database")
        tables = get_query_tables(query)
        if len(tables) != 1:
            raise Exception("Only one table may be in query to log database")
        table = tables[0]
        if table[0] is not None and table[0].lower() != 'log':
            raise Exception("This is not a query to the log database")
        if table[1].lower() not in self._tables.keys():
            raise Exception(f"There is no table '{table[1]}' in the log database")

        log_table = self._tables[table[1].lower()]

        # region check that only allowed identifiers are used in the query
        available_columns_names = [column.lower() for column in log_table.columns]

        def check_columns(node, is_table, **kwargs):
            # region replace * to available columns
            if type(node) is Select:
                new_targets = []
                for target in node.targets:
                    if type(target) is Star:
                        new_targets += [Identifier(name) for name in available_columns_names]
                    else:
                        new_targets.append(target)
                node.targets = new_targets
            # endregion

            if type(node) is Identifier and is_table is False:
                parts = resolve_table_identifier(node)
                if parts[0] is not None and parts[0].lower() not in self._tables:
                    raise Exception(f"Table '{parts[0]}' can not be used in query")
                if parts[1].lower() not in available_columns_names:
                    raise Exception(f"Column '{parts[1]}' can not be used in query")

        query_traversal(query, check_columns)
        # endregion

        query.from_table = log_table._get_base_subquery()

        render_engine = db.engine.name
        if render_engine == "postgresql":
            'postgres'
        render = SqlalchemyRender(render_engine)
        query_str = render.get_string(query, with_failback=False)
        df = pd.read_sql_query(query_str, db.engine)

        # region cast columns values to proper types
        for column_name, column_type in log_table.types_map.items():
            for df_column_name in df.columns:
                if df_column_name.lower() == column_name.lower() and df[df_column_name].dtype != column_type:
                    df[df_column_name] = df[df_column_name].astype(column_type)
        # endregion

        if return_as != 'split':
            return df

        columns_info = [{
            'name': k,
            'type': v
        } for k, v in df.dtypes.items()]

        return df, columns_info



from typing import Text, Dict, Optional

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import (
    ConnectionError,
    AuthenticationException,
    TransportError,
    RequestError,
)
from es.elastic.sqlalchemy import ESDialect
from pandas import DataFrame
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
    RESPONSE_TYPE,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class ElasticsearchHandler(DatabaseHandler):
    """
    This handler handles the connection and execution of SQL statements on Elasticsearch.
    """

    name = "elasticsearch"

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the AWS (S3) account.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self) -> None:
        """
        Closes the connection when the handler instance is deleted.
        """
        if self.is_connected:
            self.disconnect()

    def connect(self) -> Elasticsearch:
        """
        Establishes a connection to the Elasticsearch host.

        Raises:
            ValueError: If the expected connection parameters are not provided.

        Returns:
            elasticsearch.Elasticsearch: A connection object to the Elasticsearch host.
        """
        if self.is_connected is True:
            return self.connection

        config = {}

        # Mandatory connection parameters.
        if ("hosts" not in self.connection_data) and (
            "cloud_id" not in self.connection_data
        ):
            raise ValueError(
                "Either the hosts or cloud_id parameter should be provided!"
            )

        # Optional/Additional connection parameters.
        optional_parameters = ["hosts", "cloud_id", "api_key"]
        for parameter in optional_parameters:
            if parameter in self.connection_data:
                if parameter == "hosts":
                    config["hosts"] = self.connection_data[parameter].split(",")
                else:
                    config[parameter] = self.connection_data[parameter]

        # Ensure that if either user or password is provided, both are provided.
        if ("user" in self.connection_data) != ("password" in self.connection_data):
            raise ValueError(
                "Both user and password should be provided if one of them is provided!"
            )

        if "user" in self.connection_data:
            config["http_auth"] = (
                self.connection_data["user"],
                self.connection_data["password"],
            )

        try:
            self.connection = Elasticsearch(
                **config,
            )
            self.is_connected = True
            return self.connection
        except ConnectionError as conn_error:
            logger.error(
                f"Connection error when connecting to Elasticsearch: {conn_error}"
            )
            raise
        except AuthenticationException as auth_error:
            logger.error(
                f"Authentication error when connecting to Elasticsearch: {auth_error}"
            )
            raise
        except Exception as unknown_error:
            logger.error(
                f"Unknown error when connecting to Elasticsearch: {unknown_error}"
            )
            raise

    def disconnect(self) -> None:
        """
        Closes the connection to the Elasticsearch host if it's currently open.
        """
        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Elasticsearch host.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()

            # Execute a simple query to test the connection.
            connection.sql.query(body={"query": "SELECT 1"})
            response.success = True
        # All exceptions are caught here to ensure that the connection is closed if an error occurs.
        except Exception as error:
            logger.error(f"Error connecting to Elasticsearch, {error}!")
            response.error_message = str(error)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a native SQL query on the Elasticsearch host and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        try:
            response = connection.sql.query(body={"query": query})
            records = response["rows"]
            columns = response["columns"]

            new_records = True
            while new_records:
                try:
                    if response["cursor"]:
                        response = connection.sql.query(
                            body={"query": query, "cursor": response["cursor"]}
                        )

                        new_records = response["rows"]
                        records = records + new_records
                except KeyError:
                    new_records = False

            column_names = [column["name"] for column in columns]
            if not records:
                null_record = [None] * len(column_names)
                records = [null_record]

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_frame=DataFrame(records, columns=column_names),
            )

        except (TransportError, RequestError) as transport_or_request_error:
            logger.error(
                f"Error running query: {query} on Elasticsearch, {transport_or_request_error}!"
            )
            response = Response(
                RESPONSE_TYPE.ERROR, error_message=str(transport_or_request_error)
            )
        except Exception as unknown_error:
            logger.error(
                f"Unknown error running query: {query} on Elasticsearch, {unknown_error}!"
            )
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(unknown_error))

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode on the Elasticsearch host and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        # TODO: Add support for other query types.
        renderer = SqlalchemyRender(ESDialect)
        query_str = renderer.get_string(query, with_failback=True)
        logger.debug(f"Executing SQL query: {query_str}")
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables (indexes) in the Elasticsearch host.

        Returns:
            Response: A response object containing a list of tables (indexes) in the Elasticsearch host.
        """
        query = """
            SHOW TABLES
        """
        result = self.native_query(query)

        df = result.data_frame

        # Remove indices that are system indices: These are indices that start with a period.
        df = df[~df["name"].str.startswith(".")]

        df = df.drop(["catalog", "kind"], axis=1)
        result.data_frame = df.rename(
            columns={"name": "table_name", "type": "table_type"}
        )

        return result

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column (field) details for a specified table (index) in the Elasticsearch host.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"""
            DESCRIBE {table_name}
        """
        result = self.native_query(query)

        df = result.data_frame
        df = df.drop("mapping", axis=1)
        result.data_frame = df.rename(
            columns={"column": "column_name", "type": "data_type"}
        )

        return result



from flask import Flask
from waitress import serve

from mindsdb.api.http.initialize import initialize_app
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.utilities.functions import init_lexer_parsers
from mindsdb.integrations.libs.ml_exec_base import process_cache

logger = log.getLogger(__name__)


def start(verbose, no_studio, app: Flask = None):
    db.init()
    init_lexer_parsers()

    if app is None:
        app = initialize_app(config, no_studio)

    port = config['api']['http']['port']
    host = config['api']['http']['host']
    server_type = config['api']['http']['server']['type']
    server_config = config['api']['http']['server']['config']

    process_cache.init()

    if server_type == "waitress":
        logger.debug("Serving HTTP app with waitress..")
        serve(
            app,
            host='*' if host in ('', '0.0.0.0') else host,
            port=port,
            **server_config
        )
    elif server_type == "flask":
        logger.debug("Serving HTTP app with flask..")
        # that will 'disable access' log in console

        app.run(debug=False, port=port, host=host, **server_config)
    elif server_type == 'gunicorn':
        try:
            from mindsdb.api.http.gunicorn_wrapper import StandaloneApplication
        except ImportError:
            logger.error(
                "Gunicorn server is not available by default. If you wish to use it, please install 'gunicorn'"
            )
            return

        def post_fork(arbiter, worker):
            db.engine.dispose()

        def before_worker_exit(arbiter, worker):
            """Latest version of gunicorn (23.0.0) calls 'join' for each child process before exiting. However this does
            not work for processes created by ProcessPoolExecutor, because they execute forever. We need to explicitly
            call 'shutdown' for such processes before exiting.
            """
            from mindsdb.integrations.libs.process_cache import process_cache
            process_cache.shutdown(wait=True)

        options = {
            'bind': f'{host}:{port}',
            'post_fork': post_fork,
            'worker_exit': before_worker_exit,
            **server_config
        }
        StandaloneApplication(app, options).run()



import pandas as pd
from mindsdb.integrations.libs.response import HandlerStatusResponse
from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.utilities.config import Config
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe
from mindsdb.utilities.security import validate_urls
from .urlcrawl_helpers import get_all_websites


class CrawlerTable(APITable):

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.config = Config()

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Selects data from the provided websites

        Args:
            query (ast.Select): Given SQL SELECT query

        Returns:
            dataframe: Dataframe containing the crawled data

        Raises:
            NotImplementedError: If the query is not supported
        """
        conditions = extract_comparison_conditions(query.where)
        urls = []
        for operator, arg1, arg2 in conditions:
            if operator == 'or':
                raise NotImplementedError('OR is not supported')
            if arg1 == 'url':
                if operator in ['=', 'in']:
                    urls = [str(arg2)] if isinstance(arg2, str) else arg2
                else:
                    raise NotImplementedError('Invalid URL format. Please provide a single URL like url = "example.com" or'
                                              'multiple URLs using the format url IN ("url1", "url2", ...)')

        if len(urls) == 0:
            raise NotImplementedError(
                'You must specify what url you want to crawl, for example: SELECT * FROM crawl WHERE url = "someurl"')

        allowed_urls = self.config.get('web_crawling_allowed_sites', [])
        if allowed_urls and not validate_urls(urls, allowed_urls):
            raise ValueError(f"The provided URL is not allowed for web crawling. Please use any of {', '.join(allowed_urls)}.")

        if query.limit is None:
            raise NotImplementedError('You must specify a LIMIT clause which defines the number of pages to crawl')

        limit = query.limit.value

        result = get_all_websites(urls, limit, html=False)
        if len(result) > limit:
            result = result[:limit]
        # filter targets
        result = project_dataframe(result, query.targets, self.get_columns())
        return result

    def get_columns(self):
        """
        Returns the columns of the crawler table
        """
        return [
            'url',
            'text_content',
            'error'
        ]


class WebHandler(APIHandler):
    """
    Web handler, handling crawling content from websites.
    """
    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        crawler = CrawlerTable(self)
        self._register_table('crawler', crawler)

    def check_connection(self) -> HandlerStatusResponse:
        """
        Checks the connection to the web handler
        @TODO: Implement a better check for the connection

        Returns:
            HandlerStatusResponse: Response containing the status of the connection. Hardcoded to True for now.
        """
        response = HandlerStatusResponse(True)
        return response



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the EdgelessDB server.',
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the EdgelessDB server.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the EdgelessDB server.',
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the EdgelessDB server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.',
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the EdgelessDB server. Must be an integer.',
    },
    ssl={'type': ARG_TYPE.BOOL, 'description': 'Set it to False to disable ssl.'},
    ssl_ca={
        'type': ARG_TYPE.PATH,
        'description': 'Path or URL of the Certificate Authority (CA) certificate file',
    },
    ssl_cert={
        'type': ARG_TYPE.PATH,
        'description': 'Path name or URL of the server public key certificate file',
    },
    ssl_key={
        'type': ARG_TYPE.PATH,
        'description': 'The path name or URL of the server private key file',
    },
)

connection_args_example = OrderedDict(
    host='127.0.0.1', port=3306, user='root', password='password', database='database'
)



from configparser import ConfigParser


class TrinoConfigProvider:

    def __init__(self, **kwargs):
        self.config_file_name = kwargs.get('config_file_name')
        self.config_parser = ConfigParser()
        self.config_parser.read(self.config_file_name)

    def get_trino_kerberos_config(self):
        return self.config_parser['KERBEROS_CONFIG']



import os
import pytest
import psycopg

from mindsdb.integrations.handlers.postgres_handler.tests.test_postgres_handler import (
    TestPostgresConnection,
    TestPostgresQuery,
    TestPostgresTables,
    TestPostgresColumns,
    TestPostgresDisconnect
)

from mindsdb.integrations.handlers.redshift_handler.redshift_handler import RedshiftHandler


HANDLER_KWARGS = {
    "connection_data": {
        "host": os.environ.get("MDB_TEST_REDSHIFT_HOST", "127.0.0.1"),
        "port": os.environ.get("MDB_TEST_REDSHIFT_PORT", "5439"),
        "user": os.environ.get("MDB_TEST_REDSHIFT_USER", "redshift"),
        "password": os.environ.get("MDB_TEST_REDSHIFT_PASSWORD", "supersecret"),
        "database": os.environ.get("MDB_TEST_REDSHIFT_DATABASE", "redshift_db_handler_test").lower()
    }
}


@pytest.fixture(scope="module")
def handler():
    seed_db()
    handler = RedshiftHandler("test_redshift_handler", **HANDLER_KWARGS)
    yield handler
    handler.disconnect()


def seed_db():
    """
    Seed the test DB by running the queries in the seed.sql file.
    """
    # Connect to the dev (default) database to run seed queries
    conn_info = HANDLER_KWARGS["connection_data"].copy()
    del conn_info["database"]
    conn_info["dbname"] = "dev"
    db = psycopg.connect(**conn_info)
    db.autocommit = True

    cursor = db.cursor()

    # Check if the database exists
    cursor.execute(f"SELECT 1 FROM pg_database WHERE datname = '{HANDLER_KWARGS['connection_data']['database']}'")
    result = cursor.fetchone()
    # If the database exists, drop it
    if result:
        cursor.execute(f"DROP DATABASE {HANDLER_KWARGS['connection_data']['database']}")
    # Create the test database
    cursor.execute(f"CREATE DATABASE {HANDLER_KWARGS['connection_data']['database']}")

    # Reconnect to the new database
    conn_info["dbname"] = HANDLER_KWARGS['connection_data']['database']
    db = psycopg.connect(**conn_info)
    db.autocommit = True
    cursor = db.cursor()

    with open("mindsdb/integrations/handlers/redshift_handler/tests/seed.sql", "r") as f:
        for line in f.readlines():
            cursor.execute(line)
    cursor.close()
    db.close()


@pytest.mark.redshift
class TestRedshiftHandlerConnection(TestPostgresConnection):
    def test_connect(self, handler):
        """
        Tests the `connect` method to ensure it connects to the Redshift warehouse and sets the `is_connected` flag to True.
        """
        super().test_connect(handler)

    def test_check_connection(self, handler):
        """
        Tests the `check_connection` method to verify that it returns a StatusResponse object and accurately reflects the connection status.
        """
        super().test_check_connection(handler)


@pytest.mark.redshift
class TestRedshiftHandlerQuery(TestPostgresQuery):
    def test_native_query_show_dbs(self, handler):
        """
        Tests the `native_query` method to ensure it returns a list of databases in the Redshift warehouse.
        """
        dbs = handler.native_query("SELECT datname FROM pg_database")
        dbs = dbs.data_frame
        assert dbs is not None, "expected to get some data, but got None"
        assert "datname" in dbs, f"Expected to get 'datname' column in response:\n{dbs}"
        dbs = list(dbs["datname"])
        expected_db = HANDLER_KWARGS["connection_data"]["database"]
        assert (
            expected_db in dbs
        ), f"expected to have {expected_db} db in response: {dbs}"

    def test_select_query(self, handler):
        """
        Tests the `query` method with SELECT query to ensure it returns the expected data from the Redshift warehouse.
        """
        super().test_select_query(handler)


@pytest.mark.redshift
class TestRedshiftHandlerTables(TestPostgresTables):
    def test_get_tables(self, handler):
        """
        Tests the `get_tables` method to confirm it returns a list of tables in the Redshift warehouse.
        """
        super().test_get_tables(handler)

    def test_create_table(self, handler):
        """
        Tests a table creation query to ensure it creates a table in the Redshift warehouse.
        """
        super().test_create_table(handler)

    def test_drop_table(self, handler):
        """
        Tests a table drop query to ensure it drops a table in the Redshift warehouse.
        """
        super().test_drop_table(handler)


@pytest.mark.redshift
class TestRedshiftHandlerColumns(TestPostgresColumns):
    def test_get_columns(self, handler):
        """
        Tests the `get_columns` method to confirm it returns a list of columns of the test table created in the Redshift warehouse.
        """
        super().test_get_columns(handler)


@pytest.mark.redshift
class TestRedshiftHandlerDisconnect(TestPostgresDisconnect):
    def test_disconnect(self, handler):
        """
        Tests the `disconnect` method to ensure it disconnects from the Redshift warehouse and sets the `is_connected` flag to False.
        """
        super().test_disconnect(handler)

    def test_check_connection(self, handler):
        """
        Tests the `check_connection` method to verify that it returns a StatusResponse object and accurately reflects the connection status.
        """
        super().test_check_connection(handler)


if __name__ == "__main__":
    pytest.main(["-m", "redshift", __file__])



import re
import datetime as dt

import sqlalchemy as sa
from sqlalchemy.sql import operators
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import aliased
from sqlalchemy.dialects import mysql, postgresql, sqlite, mssql, oracle
from sqlalchemy.schema import CreateTable, DropTable
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql import functions as sa_fnc

from mindsdb_sql_parser import ast


sa_type_names = [
    key for key, val in sa.types.__dict__.items() if hasattr(val, '__module__')
    and val.__module__ in ('sqlalchemy.sql.sqltypes', 'sqlalchemy.sql.type_api')
]

types_map = {}
for type_name in sa_type_names:
    types_map[type_name.upper()] = getattr(sa.types, type_name)
types_map['BOOL'] = types_map['BOOLEAN']


class RenderError(Exception):
    ...


# https://github.com/sqlalchemy/sqlalchemy/discussions/9483?sort=old#discussioncomment-5312979
class INTERVAL(ColumnElement):
    def __init__(self, info):
        self.info = info
        self.type = sa.Interval()


@compiles(INTERVAL)
def _compile_interval(element, compiler, **kw):
    items = element.info.split(' ', maxsplit=1)
    if compiler.dialect.driver in ['snowflake']:
        # quote all
        args = " ".join(map(str, items))
        args = f"'{args}'"
    else:
        # quote first element
        items[0] = f"'{items[0]}'"
        args = " ".join(items)
    return "INTERVAL " + args


class SqlalchemyRender:

    def __init__(self, dialect_name):
        dialects = {
            'mysql': mysql,
            'postgresql': postgresql,
            'postgres': postgresql,
            'sqlite': sqlite,
            'mssql': mssql,
            'oracle': oracle,
            'Snowflake': oracle,
        }

        if isinstance(dialect_name, str):
            dialect = dialects[dialect_name].dialect
        else:
            dialect = dialect_name

        # remove double percent signs
        # https://docs.sqlalchemy.org/en/14/faq/sqlexpressions.html#why-are-percent-signs-being-doubled-up-when-stringifying-sql-statements
        self.dialect = dialect(paramstyle="named")
        self.dialect.div_is_floordiv = False

        if dialect_name == 'mssql':
            # update version to MS_2008_VERSION for supports_multivalues_insert
            self.dialect.server_version_info = (10,)
            self.dialect._setup_version_attributes()
        elif dialect_name == 'mysql':
            # update version for support float cast
            self.dialect.server_version_info = (8, 0, 17)

    def to_column(self, parts):
        # because sqlalchemy doesn't allow columns consist from parts therefore we do it manually

        parts2 = []

        for i in parts:
            if isinstance(i, ast.Star):
                part = '*'
            else:
                part = str(sa.column(i).compile(dialect=self.dialect))

                if not i.islower():
                    # if lower value is not be quoted
                    #   then it is quoted only because of mixed case
                    #   in that case use origin string

                    part_lower = str(sa.column(i.lower()).compile(dialect=self.dialect))
                    if part.lower() != part_lower:
                        part = i

            parts2.append(part)

        return sa.column('.'.join(parts2), is_literal=True)

    def get_alias(self, alias):
        if alias is None or len(alias.parts) == 0:
            return None
        if len(alias.parts) > 1:
            raise NotImplementedError(f'Multiple alias {alias.parts}')
        return alias.parts[0]

    def to_expression(self, t):

        # simple type
        if (
                isinstance(t, str)
                or isinstance(t, int)
                or isinstance(t, float)
                or t is None
        ):
            t = ast.Constant(t)

        if isinstance(t, ast.Star):
            col = sa.text('*')
        elif isinstance(t, ast.Last):
            col = self.to_column(['last'])
        elif isinstance(t, ast.Constant):
            col = sa.literal(t.value)
            if t.alias:
                alias = self.get_alias(t.alias)
            else:
                if t.value is None:
                    alias = 'NULL'
                else:
                    alias = str(t.value)
            col = col.label(alias)
        elif isinstance(t, ast.Identifier):
            # sql functions
            col = None
            if len(t.parts) == 1:
                if isinstance(t.parts[0], str):
                    name = t.parts[0].upper()
                    if name == 'CURRENT_DATE':
                        col = sa_fnc.current_date()
                    elif name == 'CURRENT_TIME':
                        col = sa_fnc.current_time()
                    elif name == 'CURRENT_TIMESTAMP':
                        col = sa_fnc.current_timestamp()
                    elif name == 'CURRENT_USER':
                        col = sa_fnc.current_user()
            if col is None:
                col = self.to_column(t.parts)
            if t.alias:
                col = col.label(self.get_alias(t.alias))
        elif isinstance(t, ast.Select):
            sub_stmt = self.prepare_select(t)
            col = sub_stmt.scalar_subquery()
            if t.alias:
                alias = self.get_alias(t.alias)
                col = col.label(alias)
        elif isinstance(t, ast.Function):
            fnc = self.to_function(t)
            if t.alias:
                alias = self.get_alias(t.alias)
            else:
                alias = str(t.op)
            col = fnc.label(alias)
        elif isinstance(t, ast.BinaryOperation):
            ops = {
                "+": operators.add,
                "-": operators.sub,
                "*": operators.mul,
                "/": operators.truediv,
                "%": operators.mod,
                "=": operators.eq,
                "!=": operators.ne,
                "<>": operators.ne,
                ">": operators.gt,
                "<": operators.lt,
                ">=": operators.ge,
                "<=": operators.le,
                "is": operators.is_,
                "is not": operators.is_not,
                "like": operators.like_op,
                "not like": operators.not_like_op,
                "in": operators.in_op,
                "not in": operators.not_in_op,
                "||": operators.concat_op,
            }
            functions = {
                "and": sa.and_,
                "or": sa.or_,
            }

            arg0 = self.to_expression(t.args[0])
            arg1 = self.to_expression(t.args[1])

            op = t.op.lower()
            if op in ('in', 'not in'):
                if isinstance(arg1, sa.sql.selectable.ColumnClause):
                    raise NotImplementedError(f'Required list argument for: {op}')

            sa_op = ops.get(op)

            if sa_op is not None:
                if isinstance(arg0, sa.TextClause):
                    # text doesn't have operate method, reverse operator
                    col = arg1.reverse_operate(sa_op, arg0)
                elif isinstance(arg1, sa.TextClause):
                    # both args are text, return text
                    col = sa.text(f'{arg0.compile(dialect=self.dialect)} {op} {arg1.compile(dialect=self.dialect)}')
                else:
                    col = arg0.operate(sa_op, arg1)

            elif t.op.lower() in functions:
                func = functions[t.op.lower()]
                col = func(arg0, arg1)
            else:
                col = arg0.op(t.op)(arg1)

            if t.alias:
                alias = self.get_alias(t.alias)
                col = col.label(alias)

        elif isinstance(t, ast.UnaryOperation):
            # not or munus
            opmap = {
                "NOT": "__invert__",
                "-": "__neg__",
            }
            arg = self.to_expression(t.args[0])

            method = opmap[t.op.upper()]
            col = getattr(arg, method)()
            if t.alias:
                alias = self.get_alias(t.alias)
                col = col.label(alias)

        elif isinstance(t, ast.BetweenOperation):
            col0 = self.to_expression(t.args[0])
            lim_down = self.to_expression(t.args[1])
            lim_up = self.to_expression(t.args[2])

            col = sa.between(col0, lim_down, lim_up)
        elif isinstance(t, ast.Interval):
            col = INTERVAL(t.args[0])
            if t.alias:
                alias = self.get_alias(t.alias)
                col = col.label(alias)

        elif isinstance(t, ast.WindowFunction):
            func = self.to_expression(t.function)

            partition = None
            if t.partition is not None:
                partition = [
                    self.to_expression(i)
                    for i in t.partition
                ]

            order_by = None
            if t.order_by is not None:
                order_by = []
                for f in t.order_by:
                    col0 = self.to_expression(f.field)
                    if f.direction == 'DESC':
                        col0 = col0.desc()
                    order_by.append(col0)

            rows, range_ = None, None
            if t.modifier is not None:
                words = t.modifier.lower().split()
                if words[1] == 'between' and words[4] == 'and':
                    # frame options
                    # rows/groups BETWEEN <> <> AND <> <>
                    # https://docs.sqlalchemy.org/en/20/core/sqlelement.html#sqlalchemy.sql.expression.over
                    items = []
                    for word1, word2 in (words[2:4], words[5:7]):
                        if word1 == 'unbounded':
                            items.append(None)
                        elif (word1, word2) == ('current', 'row'):
                            items.append(0)
                        elif word1.isdigits():
                            val = int(word1)
                            if word2 == 'preceding':
                                val = -val
                            elif word2 != 'following':
                                continue
                            items.append(val)
                    if len(items) == 2:
                        if words[0] == 'rows':
                            rows = tuple(items)
                        elif words[0] == 'range':
                            range_ = tuple(items)

            col = sa.over(
                func,
                partition_by=partition,
                order_by=order_by,
                range_=range_,
                rows=rows
            )

            if t.alias:
                col = col.label(self.get_alias(t.alias))
        elif isinstance(t, ast.TypeCast):
            arg = self.to_expression(t.arg)
            type = self.get_type(t.type_name)
            if t.precision is not None:
                type = type(*t.precision)
            col = sa.cast(arg, type)

            if t.alias:
                alias = self.get_alias(t.alias)
                col = col.label(alias)
        elif isinstance(t, ast.Parameter):
            col = sa.column(t.value, is_literal=True)
            if t.alias:
                raise RenderError()
        elif isinstance(t, ast.Tuple):
            col = [
                self.to_expression(i)
                for i in t.items
            ]
        elif isinstance(t, ast.Variable):
            col = sa.column(t.to_string(), is_literal=True)
        elif isinstance(t, ast.Latest):
            col = sa.column(t.to_string(), is_literal=True)
        elif isinstance(t, ast.Exists):
            sub_stmt = self.prepare_select(t.query)
            col = sub_stmt.exists()
        elif isinstance(t, ast.NotExists):
            sub_stmt = self.prepare_select(t.query)
            col = ~sub_stmt.exists()
        elif isinstance(t, ast.Case):
            col = self.prepare_case(t)
        else:
            # some other complex object?
            raise NotImplementedError(f'Column {t}')

        return col

    def prepare_case(self, t: ast.Case):
        conditions = []
        for condition, result in t.rules:
            conditions.append(
                (self.to_expression(condition), self.to_expression(result))
            )
        default = None
        if t.default is not None:
            default = self.to_expression(t.default)

        value = None
        if t.arg is not None:
            value = self.to_expression(t.arg)

        col = sa.case(*conditions, else_=default, value=value)
        if t.alias:
            col = col.label(self.get_alias(t.alias))
        return col

    def to_function(self, t):
        op = getattr(sa.func, t.op)
        if t.from_arg is not None:
            arg = t.args[0].to_string()
            from_arg = self.to_expression(t.from_arg)

            fnc = op(arg, from_arg)
        else:
            args = [
                self.to_expression(i)
                for i in t.args
            ]
            if t.distinct:
                # set first argument to distinct
                args[0] = args[0].distinct()
            fnc = op(*args)
        return fnc

    def get_type(self, typename):
        # TODO how to get type
        if not isinstance(typename, str):
            # sqlalchemy type
            return typename

        typename = typename.upper()
        if re.match(r'^INT[\d]*$', typename):
            typename = 'BIGINT'
        if re.match(r'^FLOAT[\d]*$', typename):
            typename = 'FLOAT'

        return types_map[typename]

    def prepare_join(self, join):
        # join tree to table list

        if isinstance(join.right, ast.Join):
            raise NotImplementedError('Wrong join AST')

        items = []

        if isinstance(join.left, ast.Join):
            # dive to next level
            items.extend(self.prepare_join(join.left))
        else:
            # this is first table
            items.append(dict(
                table=join.left
            ))

        # all properties set to right table
        items.append(dict(
            table=join.right,
            join_type=join.join_type,
            is_implicit=join.implicit,
            condition=join.condition
        ))

        return items

    def get_table_name(self, table_name):
        schema = None
        if isinstance(table_name, ast.Identifier):
            parts = table_name.parts

            if len(parts) > 2:
                # TODO tests is failing
                raise NotImplementedError(f'Path to long: {table_name.parts}')

            if len(parts) == 2:
                schema = parts[-2]

            table_name = parts[-1]

        return schema, table_name

    def to_table(self, node):
        if isinstance(node, ast.Identifier):
            schema, table_name = self.get_table_name(node)

            table = sa.table(table_name, schema=schema)

            if node.alias:
                table = aliased(table, name=self.get_alias(node.alias))

        elif isinstance(node, (ast.Select, ast.Union, ast.Intersect, ast.Except)):
            sub_stmt = self.prepare_select(node)
            alias = None
            if node.alias:
                alias = self.get_alias(node.alias)
            table = sub_stmt.subquery(alias)

        else:
            # TODO tests are failing
            raise NotImplementedError(f'Table {node.__name__}')

        return table

    def prepare_select(self, node):
        if isinstance(node, (ast.Union, ast.Except, ast.Intersect)):
            return self.prepare_union(node)

        cols = []
        for t in node.targets:
            col = self.to_expression(t)
            cols.append(col)

        query = sa.select(*cols)

        if node.cte is not None:
            for cte in node.cte:
                if cte.columns is not None and len(cte.columns) > 0:
                    raise NotImplementedError('CTE columns')

                stmt = self.prepare_select(cte.query)
                alias = cte.name

                query = query.add_cte(stmt.cte(self.get_alias(alias), nesting=True))

        if node.distinct:
            query = query.distinct()

        if node.from_table is not None:
            from_table = node.from_table

            if isinstance(from_table, ast.Join):
                join_list = self.prepare_join(from_table)
                # first table
                table = self.to_table(join_list[0]['table'])
                query = query.select_from(table)

                # other tables
                for item in join_list[1:]:
                    table = self.to_table(item['table'])
                    if item['is_implicit']:
                        # add to from clause
                        query = query.select_from(table)
                    else:
                        if item['condition'] is None:
                            # otherwise, sqlalchemy raises "Don't know how to join to ..."
                            condition = sa.text('1=1')
                        else:
                            condition = self.to_expression(item['condition'])

                        join_type = item['join_type']
                        method = 'join'
                        is_full = False
                        if join_type == 'LEFT JOIN':
                            method = 'outerjoin'
                        if join_type == 'FULL JOIN':
                            is_full = True

                        # perform join
                        query = getattr(query, method)(
                            table,
                            condition,
                            full=is_full
                        )
            elif isinstance(from_table, ast.Union):
                alias = None
                if from_table.alias:
                    alias = self.get_alias(from_table.alias)
                table = self.prepare_union(from_table).subquery(alias)
                query = query.select_from(table)

            elif isinstance(from_table, ast.Select):
                table = self.to_table(from_table)
                query = query.select_from(table)

            elif isinstance(from_table, ast.Identifier):
                table = self.to_table(from_table)
                query = query.select_from(table)

            elif isinstance(from_table, ast.NativeQuery):
                alias = None
                if from_table.alias:
                    alias = from_table.alias.parts[-1]
                table = sa.text(from_table.query).columns().subquery(alias)
                query = query.select_from(table)
            else:
                raise NotImplementedError(f'Select from {from_table}')

        if node.where is not None:
            query = query.filter(
                self.to_expression(node.where)
            )

        if node.group_by is not None:
            cols = [
                self.to_expression(i)
                for i in node.group_by
            ]
            query = query.group_by(*cols)

        if node.having is not None:
            query = query.having(self.to_expression(node.having))

        if node.order_by is not None:
            order_by = []
            for f in node.order_by:
                col0 = self.to_expression(f.field)
                if f.direction.upper() == 'DESC':
                    col0 = col0.desc()
                elif f.direction.upper() == 'ASC':
                    col0 = col0.asc()
                if f.nulls.upper() == 'NULLS FIRST':
                    col0 = sa.nullsfirst(col0)
                elif f.nulls.upper() == 'NULLS LAST':
                    col0 = sa.nullslast(col0)
                order_by.append(col0)

            query = query.order_by(*order_by)

        if node.limit is not None:
            query = query.limit(node.limit.value)

        if node.offset is not None:
            query = query.offset(node.offset.value)

        if node.mode is not None:
            if node.mode == 'FOR UPDATE':
                query = query.with_for_update()
            else:
                raise NotImplementedError(f'Select mode: {node.mode}')

        return query

    def prepare_union(self, from_table):
        step1 = self.prepare_select(from_table.left)
        step2 = self.prepare_select(from_table.right)

        if isinstance(from_table, ast.Except):
            func = sa.except_ if from_table.unique else sa.except_all
        elif isinstance(from_table, ast.Intersect):
            func = sa.intersect if from_table.unique else sa.intersect_all
        else:
            func = sa.union if from_table.unique else sa.union_all

        return func(step1, step2)

    def prepare_create_table(self, ast_query):
        columns = []

        for col in ast_query.columns:
            default = None
            if col.default is not None:
                if isinstance(col.default, str):
                    default = sa.text(col.default)

            if isinstance(col.type, str) and col.type.lower() == 'serial':
                col.is_primary_key = True
                col.type = 'INT'

            kwargs = {
                'primary_key': col.is_primary_key,
                'server_default': default,
            }
            if col.nullable is not None:
                kwargs['nullable'] = col.nullable

            columns.append(
                sa.Column(
                    col.name,
                    self.get_type(col.type),
                    **kwargs
                )
            )

        schema, table_name = self.get_table_name(ast_query.name)

        metadata = sa.MetaData()
        table = sa.Table(
            table_name,
            metadata,
            schema=schema,
            *columns
        )

        return CreateTable(table)

    def prepare_drop_table(self, ast_query):
        if len(ast_query.tables) != 1:
            raise NotImplementedError('Only one table is supported')

        schema, table_name = self.get_table_name(ast_query.tables[0])

        metadata = sa.MetaData()
        table = sa.Table(
            table_name,
            metadata,
            schema=schema
        )
        return DropTable(table, if_exists=ast_query.if_exists)

    def prepare_insert(self, ast_query, with_params=False):
        params = None
        schema, table_name = self.get_table_name(ast_query.table)

        names = []
        columns = []

        if ast_query.columns is None:
            raise NotImplementedError('Columns is required in insert query')
        for col in ast_query.columns:
            columns.append(
                sa.Column(
                    col.name,
                    # self.get_type(col.type)
                )
            )
            # check doubles
            if col.name in names:
                raise RenderError(f'Columns name double: {col.name}')
            names.append(col.name)

        table = sa.table(table_name, schema=schema, *columns)

        if ast_query.values is not None:
            values = []

            if ast_query.is_plain and with_params:

                for i in range(len(ast_query.columns)):
                    values.append(sa.column('%s', is_literal=True))

                values = [values]
                params = ast_query.values
            else:

                for row in ast_query.values:
                    row = [
                        self.to_expression(val)
                        for val in row
                    ]
                    values.append(row)

            stmt = table.insert().values(values)
        else:
            # is insert from subselect
            subquery = self.prepare_select(ast_query.from_select)
            stmt = table.insert().from_select(names, subquery)

        return stmt, params

    def prepare_update(self, ast_query):
        if ast_query.from_select is not None:
            raise NotImplementedError('Render of update with sub-select is not implemented')

        schema, table_name = self.get_table_name(ast_query.table)

        columns = []

        to_update = {}
        for col, value in ast_query.update_columns.items():
            columns.append(
                sa.Column(
                    col,
                )
            )

            to_update[col] = self.to_expression(value)

        table = sa.table(table_name, schema=schema, *columns)

        stmt = table.update().values(**to_update)

        if ast_query.where is not None:
            stmt = stmt.where(self.to_expression(ast_query.where))

        return stmt

    def prepare_delete(self, ast_query: ast.Delete):
        schema, table_name = self.get_table_name(ast_query.table)

        columns = []

        table = sa.table(table_name, schema=schema, *columns)

        stmt = table.delete()

        if ast_query.where is not None:
            stmt = stmt.where(self.to_expression(ast_query.where))

        return stmt

    def get_query(self, ast_query, with_params=False):
        params = None
        if isinstance(ast_query, (ast.Select, ast.Union, ast.Except, ast.Intersect)):
            stmt = self.prepare_select(ast_query)
        elif isinstance(ast_query, ast.Insert):
            stmt, params = self.prepare_insert(ast_query, with_params=with_params)
        elif isinstance(ast_query, ast.Update):
            stmt = self.prepare_update(ast_query)
        elif isinstance(ast_query, ast.Delete):
            stmt = self.prepare_delete(ast_query)
        elif isinstance(ast_query, ast.CreateTable):
            stmt = self.prepare_create_table(ast_query)
        elif isinstance(ast_query, ast.DropTables):
            stmt = self.prepare_drop_table(ast_query)
        else:
            raise NotImplementedError(f'Unknown statement: {ast_query.__class__.__name__}')
        return stmt, params

    def get_string(self, ast_query, with_failback=True):
        """
        Render query to sql string

        :param ast_query: query to render
        :param with_failback:  switch to standard render in case of error
        :return:
        """
        sql, _ = self.get_exec_params(ast_query, with_failback=with_failback, with_params=False)
        return sql

    def get_exec_params(self, ast_query, with_failback=True, with_params=True):
        """
        Render query with separated parameters and placeholders
        :param ast_query: query to render
        :param with_failback: switch to standard render in case of error
        :return: sql query and parameters
        """

        if isinstance(ast_query, (ast.CreateTable, ast.DropTables)):
            render_func = render_ddl_query
        else:
            render_func = render_dml_query

        try:
            stmt, params = self.get_query(ast_query, with_params=with_params)

            sql = render_func(stmt, self.dialect)

            return sql, params

        except (SQLAlchemyError, NotImplementedError) as e:
            if not with_failback:
                raise e

            sql_query = str(ast_query)
            if self.dialect.name == 'postgresql':
                sql_query = sql_query.replace('`', '')
            return sql_query, None


def render_dml_query(statement, dialect):

    class LiteralCompiler(dialect.statement_compiler):

        def render_literal_value(self, value, type_):

            if isinstance(value, (str, dt.date, dt.datetime, dt.timedelta)):
                return "'{}'".format(str(value).replace("'", "''"))

            return super(LiteralCompiler, self).render_literal_value(value, type_)

    return str(LiteralCompiler(dialect, statement, compile_kwargs={'literal_binds': True}))


def render_ddl_query(statement, dialect):
    class LiteralCompiler(dialect.ddl_compiler):

        def render_literal_value(self, value, type_):
            if isinstance(value, (str, dt.date, dt.datetime, dt.timedelta)):
                return "'{}'".format(str(value).replace("'", "''"))

            return super(LiteralCompiler, self).render_literal_value(value, type_)

    return str(LiteralCompiler(dialect, statement, compile_kwargs={'literal_binds': True}))



"""predictor_index

Revision ID: 4b3c9d63e89c
Revises: c67822e96833
Create Date: 2023-12-25 20:50:08.275299

"""
from alembic import op
import sqlalchemy as sa  # noqa


# revision identifiers, used by Alembic.
revision = '4b3c9d63e89c'
down_revision = 'c67822e96833'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('agent_skills', schema=None) as batch_op:
        try:
            batch_op.create_foreign_key('agent_skills_agent_id_fk', 'agents', ['agent_id'], ['id'])
            batch_op.create_foreign_key('agent_skills_skill_id_fk', 'skills', ['skill_id'], ['id'])
        except Exception:
            pass

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_index(
            'predictor_index',
            ['company_id', 'name', 'version', 'active', 'deleted_at'],
            unique=True
        )


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_index('predictor_index')

    with op.batch_alter_table('agent_skills', schema=None) as batch_op:
        try:
            batch_op.drop_constraint('agent_skills_agent_id_fk', type_='foreignkey')
            batch_op.drop_constraint('agent_skills_skill_id_fk', type_='foreignkey')
        except Exception:
            pass



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the TDEngine server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the TDEngine server.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the TDEngine server.'
    },
    url={
        'type': ARG_TYPE.STR,
        'description': 'The url of the TDEngine server Instance. '
    },
    token={
        'type': ARG_TYPE.INT,
        'description': 'Unique Token to COnnect TDEngine'
    },
)

connection_args_example = OrderedDict(
    url='127.0.0.1:6041',
    token='<token comes here>',
    user='root',
    password='taosdata',
    database='test'
)



import os
import copy
import shutil
import datetime
import tempfile
from pathlib import Path
from http import HTTPStatus

from dateutil.tz import tzlocal
from dateutil.parser import parse as parse_datetime
from flask import request
from flask_restx import Resource
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.config import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities import log
from mindsdb.utilities.functions import decrypt, encrypt
from mindsdb.utilities.log_controller import get_logs
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.response import HandlerStatusResponse


logger = log.getLogger(__name__)


@ns_conf.route('/logs')
@ns_conf.param('name', 'Get logs')
class GetLogs(Resource):
    @ns_conf.doc('get_integrations')
    @api_endpoint_metrics('GET', '/config/logs')
    def get(self):
        min_timestamp = parse_datetime(request.args['min_timestamp'])
        max_timestamp = request.args.get('max_timestamp', None)
        context = request.args.get('context', None)
        level = request.args.get('level', None)
        log_from = request.args.get('log_from', None)
        limit = request.args.get('limit', None)

        logs = get_logs(min_timestamp, max_timestamp, context, level, log_from, limit)
        return {'data': logs}


@ns_conf.route('/')
@ns_conf.param('name', 'Get config')
class GetConfig(Resource):
    @ns_conf.doc('get_config')
    @api_endpoint_metrics('GET', '/config')
    def get(self):
        config = Config()
        return {
            'auth': {
                'http_auth_enabled': config['auth']['http_auth_enabled']
            }
        }

    @ns_conf.doc('put_config')
    @api_endpoint_metrics('PUT', '/config')
    def put(self):
        data = request.json

        unknown_argumens = list(set(data.keys()) - {'auth'})
        if len(unknown_argumens) > 0:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong arguments',
                f'Unknown argumens: {unknown_argumens}'
            )

        for key in data.keys():
            unknown_argumens = list(
                set(data[key].keys()) - set(Config()[key].keys())
            )
            if len(unknown_argumens) > 0:
                return http_error(
                    HTTPStatus.BAD_REQUEST, 'Wrong arguments',
                    f'Unknown argumens: {unknown_argumens}'
                )

        Config().update(data)

        return '', 200


@ns_conf.route('/integrations')
@ns_conf.param('name', 'List all database integration')
class ListIntegration(Resource):
    @api_endpoint_metrics('GET', '/config/integrations')
    def get(self):
        return {
            'integrations': [k for k in ca.integration_controller.get_all(show_secrets=False)]
        }


@ns_conf.route('/all_integrations')
@ns_conf.param('name', 'List all database integration')
class AllIntegration(Resource):
    @ns_conf.doc('get_all_integrations')
    @api_endpoint_metrics('GET', '/config/all_integrations')
    def get(self):
        integrations = ca.integration_controller.get_all(show_secrets=False)
        return integrations


@ns_conf.route('/integrations/<name>')
@ns_conf.param('name', 'Database integration')
class Integration(Resource):
    @ns_conf.doc('get_integration')
    @api_endpoint_metrics('GET', '/config/integrations/integration')
    def get(self, name):
        integration = ca.integration_controller.get(name, show_secrets=False)
        if integration is None:
            return http_error(HTTPStatus.NOT_FOUND, 'Not found', f'Can\'t find integration: {name}')
        integration = copy.deepcopy(integration)
        return integration

    @ns_conf.doc('put_integration')
    @api_endpoint_metrics('PUT', '/config/integrations/integration')
    def put(self, name):
        params = {}
        if request.is_json:
            params.update((request.json or {}).get('params', {}))
        else:
            params.update(request.form or {})

        if len(params) == 0:
            return http_error(HTTPStatus.BAD_REQUEST, 'Wrong argument', "type of 'params' must be dict")

        files = request.files
        temp_dir = None
        if files is not None and len(files) > 0:
            temp_dir = tempfile.mkdtemp(prefix='integration_files_')
            for key, file in files.items():
                temp_dir_path = Path(temp_dir)
                file_name = Path(file.filename)
                file_path = temp_dir_path.joinpath(file_name).resolve()
                if temp_dir_path not in file_path.parents:
                    raise Exception(f'Can not save file at path: {file_path}')
                file.save(file_path)
                params[key] = str(file_path)

        is_test = params.get('test', False)
        # TODO: Move this to new Endpoint

        config = Config()
        secret_key = config.get('secret_key', 'dummy-key')

        if is_test:
            del params['test']
            handler_type = params.pop('type', None)
            params.pop('publish', None)
            try:
                handler = ca.integration_controller.create_tmp_handler(name, handler_type, params)
                status = handler.check_connection()
            except ImportError as e:
                status = HandlerStatusResponse(success=False, error_message=str(e))
            if temp_dir is not None:
                shutil.rmtree(temp_dir)

            resp = status.to_json()

            if status.success and 'code' in params:
                if hasattr(handler, 'handler_storage'):
                    # attach storage if exists
                    export = handler.handler_storage.export_files()
                    if export:
                        # encrypt with flask secret key
                        encrypted = encrypt(export, secret_key)
                        resp['storage'] = encrypted.decode()

            return resp, 200

        config = Config()
        secret_key = config.get('secret_key', 'dummy-key')

        integration = ca.integration_controller.get(name, show_secrets=False)
        if integration is not None:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                f"Integration with name '{name}' already exists"
            )

        try:
            engine = params['type']
            if engine is not None:
                del params['type']
            params.pop('publish', False)
            storage = params.pop('storage', None)
            ca.integration_controller.add(name, engine, params)

            # copy storage
            if storage is not None:
                handler = ca.integration_controller.get_data_handler(name)

                export = decrypt(storage.encode(), secret_key)
                handler.handler_storage.import_files(export)

        except Exception as e:
            logger.error(str(e))
            if temp_dir is not None:
                shutil.rmtree(temp_dir)
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, 'Error',
                f'Error during config update: {str(e)}'
            )

        if temp_dir is not None:
            shutil.rmtree(temp_dir)
        return {}, 200

    @ns_conf.doc('delete_integration')
    @api_endpoint_metrics('DELETE', '/config/integrations/integration')
    def delete(self, name):
        integration = ca.integration_controller.get(name)
        if integration is None:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Integration does not exists',
                f"Nothing to delete. '{name}' not exists."
            )
        try:
            ca.integration_controller.delete(name)
        except Exception as e:
            logger.error(str(e))
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, 'Error',
                f"Error during integration delete: {str(e)}"
            )
        return "", 200

    @ns_conf.doc('modify_integration')
    @api_endpoint_metrics('POST', '/config/integrations/integration')
    def post(self, name):
        params = {}
        params.update((request.json or {}).get('params', {}))
        params.update(request.form or {})

        if not isinstance(params, dict):
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                "type of 'params' must be dict"
            )
        integration = ca.integration_controller.get(name)
        if integration is None:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Integration does not exists',
                f"Nothin to modify. '{name}' not exists."
            )
        try:
            if 'enabled' in params:
                params['publish'] = params['enabled']
                del params['enabled']
            ca.integration_controller.modify(name, params)

        except Exception as e:
            logger.error(str(e))
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, 'Error',
                f"Error during integration modification: {str(e)}"
            )
        return "", 200


@ns_conf.route('/vars')
class Vars(Resource):
    @api_endpoint_metrics('GET', '/config/vars')
    def get(self):
        if os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false']:
            telemtry = False
        else:
            telemtry = True

        if ca.config_obj.get('disable_mongo', False):
            mongo = False
        else:
            mongo = True

        cloud = ca.config_obj.get('cloud', False)
        local_time = datetime.datetime.now(tzlocal())
        local_timezone = local_time.tzname()

        return {
            'mongo': mongo,
            'telemtry': telemtry,
            'cloud': cloud,
            'timezone': local_timezone,
        }



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class ErrPacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def setup(self):
        err_code = 0
        if 'err_code' in self._kwargs:
            err_code = self._kwargs['err_code']

        msg = 'ERROR'
        if 'msg' in self._kwargs:
            msg = self._kwargs['msg']

        self.err_header = Datum('int<1>', 255)
        self.err_code = Datum('int<2>', err_code)
        self.msg = Datum('string<EOF>', msg)

    @property
    def body(self):

        order = [
            'err_header',
            'err_code',
            'msg'
        ]
        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(str(ErrPacket().get_packet_string()))


# only run the test if this file is called from debugger
if __name__ == "__main__":
    ErrPacket.test()



"""fix_unique_constraint

Revision ID: 459218b0844c
Revises: d429095b570f
Create Date: 2022-12-26 13:40:57.141241

"""
from alembic import op

revision = '459218b0844c'
down_revision = 'd429095b570f'
branch_labels = None
depends_on = None


def upgrade():

    # try - for sqlite database
    try:
        op.execute("ALTER TABLE project DROP CONSTRAINT IF EXISTS unique_integration_name_company_id")
        op.execute("ALTER TABLE project DROP CONSTRAINT IF EXISTS unique_project_name_company_id")
    except Exception:
        pass

    try:
        with op.batch_alter_table('project', schema=None) as batch_op:
            batch_op.create_unique_constraint('unique_project_name_company_id', ['name', 'company_id'])
    except Exception:
        pass


def downgrade():
    # do nothing
    ...



"""agent_model_optional

Revision ID: bfc6f44f5bc9
Revises: 2958416fbe75
Create Date: 2024-07-09 14:50:40.193842

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = 'bfc6f44f5bc9'
down_revision = 'cbedc4968d5d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('agents', schema=None) as batch_op:
        batch_op.alter_column(
            'model_name',
            existing_type=sa.VARCHAR(),
            nullable=True
        )

    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.alter_column(
            'model_name',
            existing_type=sa.VARCHAR(),
            nullable=True
        )


def downgrade():

    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.alter_column(
            'model_name',
            existing_type=sa.VARCHAR(),
            nullable=False
        )

    with op.batch_alter_table('agents', schema=None) as batch_op:
        batch_op.alter_column(
            'model_name',
            existing_type=sa.VARCHAR(),
            nullable=False
        )



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    db_file={
        'type': ARG_TYPE.STR,
        'description': 'The database file where the data will be stored. The special path name :memory: can be provided'
                       ' to create a temporary database in RAM.'
    }
)

connection_args_example = OrderedDict(
    db_file='chinook.db'
)



"""jobs_history_query

Revision ID: b8be148dbc85
Revises: ef04cdbe51ed
Create Date: 2023-04-11 17:35:16.273293

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = 'b8be148dbc85'
down_revision = 'ef04cdbe51ed'
branch_labels = None
depends_on = None


def upgrade():

    with op.batch_alter_table('jobs_history', schema=None) as batch_op:
        batch_op.add_column(sa.Column('query_str', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('updated_at', sa.DateTime(), nullable=True,
                                      server_default=sa.func.current_timestamp()))

    # ### end Alembic commands ###


def downgrade():

    with op.batch_alter_table('jobs_history', schema=None) as batch_op:
        batch_op.drop_column('updated_at')
        batch_op.drop_column('query_str')

    # ### end Alembic commands ###



from pathlib import Path

import pandas as pd

DATASETS_BASE_PATH = Path(__file__).parent

SUPPORTED_TASK_TYPES = ("question_answering",)


class DatasetNameMissing(Exception):
    pass


class MLTaskTypeMissing(Exception):
    pass


class DatasetNotFound(Exception):
    pass


class UnsupportedMLTaskType(Exception):
    pass


class MissingColumns(Exception):
    pass


def validate_dataset(ml_task_type=None, dataset_name=None):

    if ml_task_type is None:
        raise MLTaskTypeMissing(
            "ML Task type is missing. Please provide a valid 'ml_task_type'."
        )

    if dataset_name is None:
        raise DatasetNameMissing(
            "Dataset name is missing. Please provide a valid 'dataset_name'."
        )

    if ml_task_type not in SUPPORTED_TASK_TYPES:
        raise UnsupportedMLTaskType(
            f"ML Task type '{ml_task_type}' is not supported. Supported types are: {SUPPORTED_TASK_TYPES}"
        )

    dataset_path = DATASETS_BASE_PATH / ml_task_type / f"{dataset_name}.csv"

    if not dataset_path.exists():
        raise DatasetNotFound(
            f"Dataset '{dataset_name}' for ML Task type '{ml_task_type}' not found '{dataset_path}'."
        )

    return dataset_path


def load_dataset(ml_task_type=None, dataset_name=None):

    dataset_path = validate_dataset(ml_task_type, dataset_name)

    return pd.read_csv(dataset_path)


def validate_dataframe(df, mandatory_columns):

    columns_exist = all([col in df.columns for col in mandatory_columns])

    if not columns_exist:
        raise MissingColumns(
            f"Columns {mandatory_columns} are missing from the dataframe."
        )

    return df



from typing import List

import pandas as pd

from mindsdb_sql_parser.ast import (
    Identifier, BinaryOperation, Last, Constant, ASTNode
)
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.interfaces.storage import db
from mindsdb.utilities.context import context as ctx

from .last_query import LastQuery


class QueryContextController:
    IGNORE_CONTEXT = '<IGNORE>'

    def handle_db_context_vars(self, query: ASTNode, dn, session) -> tuple:
        """
        Check context variables in query and replace them with values.
        Should be used before exec query in database.

        Input:
        - query: input query
        - params are used to find current values of context variables
          - dn: datanode
          - session: mindsdb server session

        Returns:
         - query with replaced context variables
         - callback to call with result of the query. it is used to update context variables

        """
        context_name = self.get_current_context()

        l_query = LastQuery(query)
        if l_query.query is None:
            # no last keyword, exit
            return query, None

        if context_name == self.IGNORE_CONTEXT:
            # return with empty constants
            return l_query.query, None

        query_str = l_query.to_string()

        rec = self._get_context_record(context_name, query_str)

        if rec is None or len(rec.values) == 0:
            values = self._get_init_last_values(l_query, dn, session)
            if rec is None:
                self.__add_context_record(context_name, query_str, values)
                if context_name.startswith('job-if-'):
                    # add context for job also
                    self.__add_context_record(context_name.replace('job-if', 'job'), query_str, values)
            else:
                rec.values = values
        else:
            values = rec.values

        db.session.commit()

        query_out = l_query.apply_values(values)

        def callback(df, columns_info):
            self._result_callback(l_query, context_name, query_str, df, columns_info)

        return query_out, callback

    def remove_lasts(self, query):
        def replace_lasts(node, **kwargs):

            # find last in where
            if isinstance(node, BinaryOperation):
                if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Last):
                    node.args = [Constant(0), Constant(0)]
                    node.op = '='

        # find lasts
        query_traversal(query, replace_lasts)
        return query

    def _result_callback(self, l_query: LastQuery,
                         context_name: str, query_str: str,
                         df: pd.DataFrame, columns_info: list):
        """
        This function handlers result from executed query and updates context variables with new values

        Input
        - l_query: LastQuery object
        - To identify context:
          - context_name: name of the context
          - query_str: rendered query to search in context table
        - result of the query
          - data: list of dicts
          - columns_info: list

        """
        if len(df) == 0:
            return

        values = {}
        # get max values
        for info in l_query.get_last_columns():
            target_idx = info['target_idx']
            if target_idx is not None:
                # get by index
                col_name = columns_info[target_idx]['name']
            else:
                col_name = info['column_name']
                # get by name
            if col_name not in df:
                continue

            column_values = df[col_name].dropna()
            try:
                value = max(column_values)
            except (TypeError, ValueError):
                try:
                    # try to convert to float
                    value = max(map(float, column_values))
                except (TypeError, ValueError):
                    try:
                        # try to convert to str
                        value = max(map(str, column_values))
                    except (TypeError, ValueError):
                        continue

            if value is not None:
                values[info['table_name']] = {info['column_name']: value}

        self.__update_context_record(context_name, query_str, values)

    def drop_query_context(self, object_type: str, object_id: int = None):
        """
        Drop context for object
        :param object_type: type of the object
        :param object_id: id
        """

        context_name = self.gen_context_name(object_type, object_id)
        for rec in db.session.query(db.QueryContext).filter_by(
            context_name=context_name,
            company_id=ctx.company_id
        ).all():
            db.session.delete(rec)
        db.session.commit()

    def _get_init_last_values(self, l_query: LastQuery, dn, session) -> dict:
        """
        Gets current last values for query.
        Creates and executes query for it:
           'select <col> from <table> order by <col> desc limit 1"
        """
        last_values = {}
        for query, info in l_query.get_init_queries():

            data, columns_info = dn.query(
                query=query,
                session=session
            )

            if len(data) == 0:
                value = None
            else:
                row = list(data.iloc[0])

                idx = None
                for i, col in enumerate(columns_info):
                    if col['name'].upper() == info['column_name'].upper():
                        idx = i
                        break

                if idx is None or len(row) == 1:
                    value = row[0]
                else:
                    value = row[idx]

            if value is not None:
                last_values[info['table_name']] = {info['column_name']: value}

        return last_values

    # Context

    def get_current_context(self) -> str:
        """
        returns current context name
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        if len(context_stack) > 0:
            return context_stack[-1]
        else:
            return ''

    def set_context(self, object_type: str = None, object_id: int = None):
        """
        Updates current context name, using object name and id
        Previous context names are stored on lower levels of stack
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        context_stack.append(self.gen_context_name(object_type, object_id))
        ctx.context_stack = context_stack

    def release_context(self, object_type: str = None, object_id: int = None):
        """
        Removed current context (defined by object type and id) and restored previous one
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        if len(context_stack) == 0:
            return
        context_name = self.gen_context_name(object_type, object_id)
        if context_stack[-1] == context_name:
            context_stack.pop()
        ctx.context_stack = context_stack

    def gen_context_name(self, object_type: str, object_id: int) -> str:
        """
        Generated name of the context according to object type and name
        :return: context name
        """

        if object_type is None:
            return ''
        if object_id is not None:
            object_type += '-' + str(object_id)
        return object_type

    def get_context_vars(self, object_type: str, object_id: int) -> List[dict]:
        """
        Return variables stored in context (defined by object type and id)

        :return: list of all context variables related to context name how they stored in context table
        """
        context_name = self.gen_context_name(object_type, object_id)
        vars = []
        for rec in db.session.query(db.QueryContext).filter_by(
            context_name=context_name,
            company_id=ctx.company_id
        ):
            if rec.values is not None:
                vars.append(rec.values)

        return vars

    # DB
    def _get_context_record(self, context_name: str, query_str: str) -> db.QueryContext:
        """
        Find and return record for context and query string
        """

        return db.session.query(db.QueryContext).filter_by(
            query=query_str,
            context_name=context_name,
            company_id=ctx.company_id
        ).first()

    def __add_context_record(self, context_name: str, query_str: str, values: dict) -> db.QueryContext:
        """
        Creates record (for context and query string) with values and returns it
        """
        rec = db.QueryContext(
            query=query_str,
            context_name=context_name,
            company_id=ctx.company_id,
            values=values)
        db.session.add(rec)
        return rec

    def __update_context_record(self, context_name: str, query_str: str, values: dict):
        """
        Updates context record with new values
        """
        rec = self._get_context_record(context_name, query_str)
        rec.values = values
        db.session.commit()


query_context_controller = QueryContextController()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    fauna_secret={
        "type": ARG_TYPE.STR,
        "description": "faunadb secret",
        "required": True,
        "secret": True
    },
    fauna_scheme={
        "type": ARG_TYPE.STR,
        "description": "faunadb scheme(http/https)",
        "required": False,
    },
    fauna_domain={
        "type": ARG_TYPE.STR,
        "description": "faunadb instance domain",
        "required": False,
    },
    fauna_port={
        "type": ARG_TYPE.INT,
        "description": "faunadb instance port",
        "required": False,
    },
    fauna_endpoint={
        "type": ARG_TYPE.STR,
        "description": "faunadb instance endpoint",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    fauna_scheme="https",
    fauna_domain="db.fauna.com",
    fauna_port=443,
)




import pandas as pd
from mindsdb_sql_parser.ast import BinaryOperation, Constant, Identifier, Select
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

from mindsdb.api.executor.datahub.classes.tables_row import (
    TABLES_ROW_TYPE,
    TablesRow,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class Table:

    deletable: bool = False
    visible: bool = False
    kind: str = 'table'


class SchemataTable(Table):
    name = 'SCHEMATA'
    columns = [
        "CATALOG_NAME",
        "SCHEMA_NAME",
        "DEFAULT_CHARACTER_SET_NAME",
        "DEFAULT_COLLATION_NAME",
        "SQL_PATH",
    ]

    @classmethod
    def get_data(cls, inf_schema=None, **kwargs):

        databases_meta = inf_schema.session.database_controller.get_list()
        data = [
            ["def", x["name"], "utf8mb4", "utf8mb4_0900_ai_ci", None]
            for x in databases_meta
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class TablesTable(Table):
    name = 'TABLES'

    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "TABLE_TYPE",
        "ENGINE",
        "VERSION",
        "ROW_FORMAT",
        "TABLE_ROWS",
        "AVG_ROW_LENGTH",
        "DATA_LENGTH",
        "MAX_DATA_LENGTH",
        "INDEX_LENGTH",
        "DATA_FREE",
        "AUTO_INCREMENT",
        "CREATE_TIME",
        "UPDATE_TIME",
        "CHECK_TIME",
        "TABLE_COLLATION",
        "CHECKSUM",
        "CREATE_OPTIONS",
        "TABLE_COMMENT",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):

        target_table = None
        if (
            type(query) is Select
            and type(query.where) is BinaryOperation
            and query.where.op == "and"
        ):
            for arg in query.where.args:
                if (
                    type(arg) is BinaryOperation
                    and arg.op == "="
                    and type(arg.args[0]) is Identifier
                    and arg.args[0].parts[-1].upper() == "TABLE_SCHEMA"
                    and type(arg.args[1]) is Constant
                ):
                    target_table = arg.args[1].value
                    break

        data = []
        for name in inf_schema.tables.keys():
            if target_table is not None and target_table != name:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in inf_schema.persis_datanodes.items():
            if target_table is not None and target_table != ds_name:
                continue
            if hasattr(ds, 'get_tables_rows'):
                ds_tables = ds.get_tables_rows()
            else:
                ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [
                    TablesRow(
                        TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x["name"]
                    )
                    for x in ds_tables
                ]
            elif (
                isinstance(ds_tables, list)
                and len(ds_tables) > 0
                and isinstance(ds_tables[0], str)
            ):
                ds_tables = [
                    TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x)
                    for x in ds_tables
                ]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in inf_schema.get_integrations_names():
            if target_table is not None and target_table != ds_name:
                continue
            try:
                ds = inf_schema.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                logger.error(f"Can't get tables from '{ds_name}'")

        for project_name in inf_schema.get_projects_names():
            if target_table is not None and target_table != project_name:
                continue
            project_dn = inf_schema.get(project_name)
            project_tables = project_dn.get_tables()
            for row in project_tables:
                row.TABLE_SCHEMA = project_name
                data.append(row.to_list())

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class ColumnsTable(Table):

    name = 'COLUMNS'
    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "ORDINAL_POSITION",
        "COLUMN_DEFAULT",
        "IS_NULLABLE",
        "DATA_TYPE",
        "CHARACTER_MAXIMUM_LENGTH",
        "CHARACTER_OCTET_LENGTH",
        "NUMERIC_PRECISION",
        "NUMERIC_SCALE",
        "DATETIME_PRECISION",
        "CHARACTER_SET_NAME",
        "COLLATION_NAME",
        "COLUMN_TYPE",
        "COLUMN_KEY",
        "EXTRA",
        "PRIVILEGES",
        "COLUMN_COMMENT",
        "GENERATION_EXPRESSION",
    ]

    @classmethod
    def get_data(cls, inf_schema=None, query: ASTNode = None, **kwargs):

        # NOTE there is a lot of types in mysql, but listed below should be enough for our purposes
        row_templates = {
            "text": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "varchar",
                1024,
                3072,
                None,
                None,
                None,
                "utf8",
                "utf8_bin",
                "varchar(1024)",
                None,
                None,
                "select",
                None,
                None,
            ],
            "timestamp": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                "CURRENT_TIMESTAMP",
                "YES",
                "timestamp",
                None,
                None,
                None,
                None,
                0,
                None,
                None,
                "timestamp",
                None,
                None,
                "select",
                None,
                None,
            ],
            "bigint": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "bigint",
                None,
                None,
                20,
                0,
                None,
                None,
                None,
                "bigint unsigned",
                None,
                None,
                "select",
                None,
                None,
            ],
            "float": [
                "def",
                "SCHEMA_NAME",
                "TABLE_NAME",
                "COLUMN_NAME",
                "COL_INDEX",
                None,
                "YES",
                "float",
                None,
                None,
                12,
                0,
                None,
                None,
                None,
                "float",
                None,
                None,
                "select",
                None,
                None,
            ],
        }

        result = []

        databases = None
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if arg1.lower() == 'table_schema':
                if op == '=':
                    databases = [arg2]
                elif op == 'in':
                    if not isinstance(arg2, list):
                        arg2 = [arg2]
                    databases = arg2

        if databases is None:
            databases = ['information_schema', 'mindsdb', 'files']

        for db_name in databases:
            tables = {}
            if db_name == 'information_schema':
                for table_name, table in inf_schema.tables.items():
                    tables[table_name] = [
                        {'name': name} for name in table.columns
                    ]
            else:
                dn = inf_schema.get(db_name)
                if dn is None:
                    continue
                for table_row in dn.get_tables():
                    tables[table_row.TABLE_NAME] = dn.get_table_columns(table_row.TABLE_NAME)

            for table_name, table_columns in tables.items():
                for i, column in enumerate(table_columns):
                    column_name = column['name']
                    column_type = column.get('type', 'text')
                    if column_type in ('double precision', 'real', 'numeric'):
                        column_type = 'float'
                    elif column_type in ('integer', 'smallint', 'int'):
                        column_type = 'bigint'
                    elif column_type in ('timestamp without time zone', 'timestamp with time zone', 'date'):
                        column_type = 'timestamp'
                    elif column_type not in row_templates:
                        column_type = 'text'
                    result_row = row_templates[column_type].copy()
                    result_row[1] = db_name
                    result_row[2] = table_name
                    result_row[3] = column_name
                    result_row[4] = i
                    result.append(result_row)

        df = pd.DataFrame(result, columns=cls.columns)
        return df


class EventsTable(Table):
    name = "EVENTS"

    columns = [
        "EVENT_CATALOG",
        "EVENT_SCHEMA",
        "EVENT_NAME",
        "DEFINER",
        "TIME_ZONE",
        "EVENT_BODY",
        "EVENT_DEFINITION",
        "EVENT_TYPE",
        "EXECUTE_AT",
        "INTERVAL_VALUE",
        "INTERVAL_FIELD",
        "SQL_MODE",
        "STARTS",
        "ENDS",
        "STATUS",
        "ON_COMPLETION",
        "CREATED",
        "LAST_ALTERED",
        "LAST_EXECUTED",
        "EVENT_COMMENT",
        "ORIGINATOR",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class RoutinesTable(Table):
    name = "ROUTINE"
    columns = [
        "SPECIFIC_NAME",
        "ROUTINE_CATALOG",
        "ROUTINE_SCHEMA",
        "ROUTINE_NAME",
        "ROUTINE_TYPE",
        "DATA_TYPE",
        "CHARACTER_MAXIMUM_LENGTH",
        "CHARACTER_OCTET_LENGTH",
        "NUMERIC_PRECISION",
        "NUMERIC_SCALE",
        "DATETIME_PRECISION",
        "CHARACTER_SET_NAME",
        "COLLATION_NAME",
        "DTD_IDENTIFIER",
        "ROUTINE_BODY",
        "ROUTINE_DEFINITION",
        "EXTERNAL_NAME",
        "EXTERNAL_LANGUAGE",
        "PARAMETER_STYLE",
        "IS_DETERMINISTIC",
        "SQL_DATA_ACCESS",
        "SQL_PATH",
        "SECURITY_TYPE",
        "CREATED",
        "LAST_ALTERED",
        "SQL_MODE",
        "ROUTINE_COMMENT",
        "DEFINER",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class PluginsTable(Table):
    name = "PLUGINS"
    columns = [
        "PLUGIN_NAME",
        "PLUGIN_VERSION",
        "PLUGIN_STATUS",
        "PLUGIN_TYPE",
        "PLUGIN_TYPE_VERSION",
        "PLUGIN_LIBRARY",
        "PLUGIN_LIBRARY_VERSION",
        "PLUGIN_AUTHOR",
        "PLUGIN_DESCRIPTION",
        "PLUGIN_LICENSE",
        "LOAD_OPTION",
        "PLUGIN_MATURITY",
        "PLUGIN_AUTH_VERSION",
    ]


class EnginesTable(Table):
    name = "ENGINES"
    columns = ["ENGINE", "SUPPORT", "COMMENT", "TRANSACTIONS", "XA", "SAVEPOINTS"]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            [
                "InnoDB",
                "DEFAULT",
                "Supports transactions, row-level locking, and foreign keys",
                "YES",
                "YES",
                "YES",
            ]
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class KeyColumnUsageTable(Table):
    name = "KEY_COLUMN_USAGE"
    columns = [
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINT_NAME",
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "ORDINAL_POSITION",
        "POSITION_IN_UNIQUE_CONSTRAINT",
        "REFERENCED_TABLE_SCHEMA",
        "REFERENCED_TABLE_NAME",
        "REFERENCED_COLUMN_NAME",
    ]


class StatisticsTable(Table):
    name = "STATISTICS"
    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "NON_UNIQUE",
        "INDEX_SCHEMA",
        "INDEX_NAME",
        "SEQ_IN_INDEX",
        "COLUMN_NAME",
        "COLLATION",
        "CARDINALITY",
        "SUB_PART",
        "PACKED",
        "NULLABLE",
        "INDEX_TYPE",
        "COMMENT",
        "INDEX_COMMENT",
        "IS_VISIBLE",
        "EXPRESSION",
    ]


class CharacterSetsTable(Table):
    name = "CHARACTER_SETS"
    columns = [
        "CHARACTER_SET_NAME",
        "DEFAULT_COLLATE_NAME",
        "DESCRIPTION",
        "MAXLEN",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8", "UTF-8 Unicode", "utf8_general_ci", 3],
            ["latin1", "cp1252 West European", "latin1_swedish_ci", 1],
            ["utf8mb4", "UTF-8 Unicode", "utf8mb4_general_ci", 4],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class CollationsTable(Table):
    name = "COLLATIONS"

    columns = [
        "COLLATION_NAME",
        "CHARACTER_SET_NAME",
        "ID",
        "IS_DEFAULT",
        "IS_COMPILED",
        "SORTLEN",
        "PAD_ATTRIBUTE",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8_general_ci", "utf8", 33, "Yes", "Yes", 1, "PAD SPACE"],
            ["latin1_swedish_ci", "latin1", 8, "Yes", "Yes", 1, "PAD SPACE"],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df



import os
import pytest
import pandas as pd

from mindsdb_sql_parser import parse_sql

from tests.unit.executor_test_base import BaseExecutorTest

class TestUnify(BaseExecutorTest):
    
    def run_sql(self, sql):
        ret = self.command_executor.execute_command(parse_sql(sql))
        assert ret.error_code is None
        if ret.data is not None:
            return ret.data.to_df()


    """
    Integration tests for the Unify handler.
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup_method(self):
        """
        Setup test environment, creating a project
        """
        super().setup_method()
        self.run_sql("create database proj")
    
    @pytest.mark.skipif(os.environ.get('UNIFY_API_KEY') is None, reason='Missing API key!')
    def test_unify_correct_flow(self):
        self.run_sql(
        f"""
        CREATE ML_ENGINE unify_engine
        FROM unify
        USING
            unify_api_key = '{os.environ.get('UNIFY_API_KEY')}';
        """
        )
        
        self.run_sql(
        f"""
        CREATE MODEL proj.test_unify_correct_flow
        PREDICT output
        USING
            engine='unify_engine',
            model = 'llama-3-8b-chat',
            provider = 'together-ai',
            question_column = 'text';
        """
        )
        self.wait_predictor("proj", "test_unify_correct_flow")

        result_df = self.run_sql(
        """
            SELECT text, output
            FROM proj.test_unify_correct_flow
            WHERE text = 'Hello';
        """
        )
        assert "hello" in result_df["output"].iloc[0].lower()



import re
from typing import Union

from langchain_core.agents import AgentAction, AgentFinish

from langchain.agents.agent import AgentOutputParser
from langchain.agents.conversational.prompt import FORMAT_INSTRUCTIONS

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SafeOutputParser(AgentOutputParser):
    '''Output parser for the conversational agent that does not throw OutputParserException.'''

    ai_prefix: str = 'AI'
    '''Prefix to use before AI output.'''

    format_instructions: str = FORMAT_INSTRUCTIONS
    '''Default formatting instructions'''

    def get_format_instructions(self) -> str:
        '''Returns formatting instructions for the given output parser.'''
        return self.format_instructions

    def parse(self, text: str) -> Union[AgentAction, AgentFinish]:
        '''Parses outputted text from an LLM.

        Args:
            text (str): Outputted text to parse.

        Returns:
            output (str): Parsed text to an Agent step.
        '''
        regex = r'Action: (.*?)[\n]*Action Input:([\s\S]*)'
        match = re.search(regex, text, re.DOTALL)
        if match is not None:
            action = match.group(1)
            action_input = match.group(2)
            return AgentAction(action.strip(), action_input.strip(' ').strip('"'), text)
        output = text
        if f'{self.ai_prefix}:' in text:
            output = text.split(f'{self.ai_prefix}:')[-1].strip()
        return AgentFinish(
            {'output': output}, text
        )

    @property
    def _type(self) -> str:
        return 'conversational'



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description
from .connection_args import connection_args
try:
    from .byom_handler import BYOMHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e


title = 'BYOM'
name = 'byom'
type = HANDLER_TYPE.ML
icon_path = "icon.svg"

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'connection_args'
]




import requests
import uuid


class ClipdropClient:
    def __init__(self, api_key, local_dir):
        self.api_key = api_key
        self.dir_to_save = local_dir if local_dir.endswith("/") else local_dir + "/"
        self.base_endpoint = "https://clipdrop-api.co/"

    def write_to_file(self, img):
        f_name = self.dir_to_save + str(uuid.uuid4()) + ".png"
        with open(f_name, "wb") as f:
            f.write(img)
        return f_name

    def make_request(self, url, files={}, data={}):
        headers = {"x-api-key": self.api_key}
        resp = requests.post(url, headers=headers, files=files, data=data)
        if resp.status_code == 200:
            saved_path = self.write_to_file(resp.content)
            return saved_path
        return resp.text

    def image_extension_check(self, url):
        ext = url.split(".")[-1]
        if ext in ["jpeg", "jpg", "png"]:
            return ext
        raise Exception("Unknown image format. Currently jpg, jpeg & png are supported.")

    def download_image(self, path):

        img_ext = self.image_extension_check(path)
        try:
            res = requests.get(path)
            return {"img_ext": img_ext, "content": res.content}
        except Exception as e:
            raise Exception(f"Failed to download image: {e}")

    def remove_text(self, img_url):
        url = f'{self.base_endpoint}remove-text/v1'
        img_content = self.download_image(img_url)
        files = {
            'image_file': ('image.jpg', img_content["content"], f'image/{img_content["img_ext"]}')
        }
        return self.make_request(url, files=files)

    def remove_background(self, img_url):
        url = f'{self.base_endpoint}remove-background/v1'
        img_content = self.download_image(img_url)
        files = {
            'image_file': ('image.jpg', img_content["content"], f'image/{img_content["img_ext"]}')
        }
        return self.make_request(url, files=files)

    def sketch_to_image(self, img_url, prompt):
        url = f'{self.base_endpoint}sketch-to-image/v1/sketch-to-image'
        img_content = self.download_image(img_url)
        files = {
            'sketch_file': ('image.jpg', img_content["content"], f'image/{img_content["img_ext"]}')
        }
        data = {'prompt': prompt}
        return self.make_request(url, files=files, data=data)

    def text_to_image(self, prompt):
        url = f'{self.base_endpoint}text-to-image/v1'
        files = {
            'prompt': (None, prompt, 'text/plain')
        }
        return self.make_request(url, files=files)

    def replace_background(self, img_url, prompt):
        url = f'{self.base_endpoint}replace-background/v1'
        img_content = self.download_image(img_url)
        files = {
            'image_file': ('image.jpg', img_content["content"], f'image/{img_content["img_ext"]}')
        }
        data = {'prompt': prompt}
        return self.make_request(url, files=files, data=data)

    def reimagine(self, img_url):
        url = f'{self.base_endpoint}reimagine/v1/reimagine'
        img_content = self.download_image(img_url)
        files = {
            'image_file': ('image.jpg', img_content["content"], f'image/{img_content["img_ext"]}')
        }
        return self.make_request(url, files=files)



import os
from typing import List, Iterator
from langchain_core.documents import Document as LangchainDocument
from langchain_text_splitters import MarkdownHeaderTextSplitter, RecursiveCharacterTextSplitter
import pandas as pd

from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.utilities.rag.loaders.file_loader import FileLoader
from mindsdb.integrations.utilities.rag.splitters.file_splitter import (
    FileSplitter,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_CHUNK_OVERLAP
)
from mindsdb.integrations.handlers.web_handler.urlcrawl_helpers import get_all_websites
from mindsdb.interfaces.knowledge_base.preprocessing.models import Document
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class DocumentLoader:
    """Handles loading documents from various sources including SQL queries"""

    def __init__(
            self,
            file_controller: FileController,
            file_splitter: FileSplitter,
            markdown_splitter: MarkdownHeaderTextSplitter,
            file_loader_class=FileLoader,
            mysql_proxy=None
    ):
        """
        Initialize with required dependencies

        Args:
            file_controller: Controller for file operations
            file_splitter: Splitter for file content
            markdown_splitter: Splitter for markdown content
            file_loader_class: Class to use for file loading
            mysql_proxy: Proxy for executing MySQL queries
        """
        self.file_controller = file_controller
        self.file_splitter = file_splitter
        self.markdown_splitter = markdown_splitter
        self.file_loader_class = file_loader_class
        self.mysql_proxy = mysql_proxy

        # Initialize text splitter for query results with default settings
        self.query_splitter = RecursiveCharacterTextSplitter(
            chunk_size=DEFAULT_CHUNK_SIZE,
            chunk_overlap=DEFAULT_CHUNK_OVERLAP
        )

    def load_files(self, file_names: List[str]) -> Iterator[Document]:
        """Load and split documents from files"""
        for file_name in file_names:
            file_path = self.file_controller.get_file_path(file_name)
            loader = self.file_loader_class(file_path)

            for doc in loader.lazy_load():
                # Add file extension to metadata for proper splitting
                extension = os.path.splitext(file_path)[1].lower()
                doc.metadata['extension'] = extension
                doc.metadata['source'] = file_name

                # Use FileSplitter to handle the document based on its type
                split_docs = self.file_splitter.split_documents([doc])
                for split_doc in split_docs:
                    # Preserve original metadata while adding split-specific metadata
                    metadata = doc.metadata.copy()
                    metadata.update(split_doc.metadata or {})

                    yield Document(
                        content=split_doc.page_content,
                        metadata=metadata
                    )

    def load_web_pages(
            self,
            urls: List[str],
            crawl_depth: int,
            limit: int,
            filters: List[str] = None,
    ) -> Iterator[Document]:
        """Load and split documents from web pages"""
        websites_df = get_all_websites(
            urls,
            crawl_depth=crawl_depth,
            limit=limit,
            filters=filters
        )

        for _, row in websites_df.iterrows():
            # Create a document with HTML extension for proper splitting
            doc = LangchainDocument(
                page_content=row['text_content'],
                metadata={
                    'extension': '.html',
                    'url': row['url']
                }
            )

            # Use FileSplitter to handle HTML content
            split_docs = self.file_splitter.split_documents([doc])
            for split_doc in split_docs:
                metadata = doc.metadata.copy()
                metadata.update(split_doc.metadata or {})

                yield Document(
                    content=split_doc.page_content,
                    metadata=metadata
                )

    def load_query_result(self, query: str, project_name: str) -> Iterator[Document]:
        """
        Load documents from SQL query results

        Args:
            query: SQL query to execute
            project_name: Name of the project context

        Returns:
            Iterator of Document objects

        Raises:
            ValueError: If mysql_proxy is not configured or query returns no data
        """
        if not self.mysql_proxy:
            raise ValueError("MySQL proxy not configured")

        if not query:
            return

        # Set project context and execute query
        self.mysql_proxy.set_context({'db': project_name})
        query_result = self.mysql_proxy.process_query(query)

        if query_result.type != 'table':
            raise ValueError('Query returned no data')

        # Convert query result to DataFrame
        df = query_result.data.to_df()

        # Process each row into a Document
        for _, row in df.iterrows():
            # Extract content and metadata
            content = str(row.get('content', ''))

            # Convert remaining columns to metadata
            metadata = {
                col: str(row[col])
                for col in df.columns
                if col != 'content' and not pd.isna(row[col])
            }
            metadata['source'] = 'query'

            # Split content using recursive splitter
            if content:
                doc = LangchainDocument(
                    page_content=content,
                    metadata=metadata
                )
                # Use FileSplitter with default recursive splitter
                split_docs = self.file_splitter.split_documents(
                    [doc],
                    default_failover=True
                )

                for split_doc in split_docs:
                    metadata = doc.metadata.copy()
                    metadata.update(split_doc.metadata or {})

                    yield Document(
                        content=split_doc.page_content,
                        metadata=metadata
                    )



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
        host={
            'type': ARG_TYPE.STR,
            'description': 'The host name or IP address of the SAP MaxDB server.'
        },
        port={
            'type': ARG_TYPE.INT,
            'description': 'The TCP/IP port of the SAP MaxDB. Must be an integer.'
        },
        database={
            'type': ARG_TYPE.STR,
            'description': 'The database name to use when connecting with the SAP MaxDB server.'
        },
        jdbc_location={
            'type': ARG_TYPE.STR,
            'description': 'The location of the jar file which contains the JDBC class.'
        },
        user={
            'type': ARG_TYPE.STR,
            'description': 'The user name used to authenticate with the SAP MaxDB server.'
        },
        password={
            'type': ARG_TYPE.PWD,
            'description': 'The password to authenticate the user with the SAP MaxDB server.',
            'secret': True
        }
)


connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=7210,
    user='DBADMIN',
    password='password',
    database="MAXDB",
    jdbc_location='/Users/marsid/Desktop/sapdbc.jar',
)



import importlib
from textwrap import dedent
from types import ModuleType
from typing import Optional, Union

from pandas import DataFrame

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.interfaces.model.model_controller import ModelController


def get_module_import_error_str(module: ModuleType) -> str:
    '''Make a str with human-readable module import error message

    Atrs:
        module (ModuleType): module with import error

    Returns:
        str: error message
    '''
    is_cloud = Config().get('cloud', False)

    msg = dedent(f'''\
        ML engine '{module.name}' cannot be used. Reason is:
            {module.import_error}
    ''')

    if is_cloud is False:
        msg += '\n'
        msg += dedent(f'''\
            If error is related to missing dependencies, then try to run command in shell and restart mindsdb:
                pip install mindsdb[{module.name}]
        ''')

    return msg


def describe_process(integration_id: int, attribute: Optional[Union[str, list]],
                     model_id: int, module_path: str) -> DataFrame:
    '''get a model description

    Args:
        model_id (int): id of the model
        integration_id (int): id of the integration
        attribute (Optional[Union[str, list]]): attribute, or list model attributes to describe
        module_path: (str): path integration module

    Returns:
        DataFrame: usually 1-row dataframe with model description
    '''
    module = importlib.import_module(module_path)

    handlerStorage = HandlerStorage(integration_id)
    modelStorage = ModelStorage(model_id)

    model_record = db.Predictor.query.get(model_id)
    if model_record is None:
        return DataFrame(['The model does not exist'], columns=['error'])

    if isinstance(attribute, str) and attribute.lower() == 'import_error':
        return DataFrame([get_module_import_error_str(module)], columns=['error'])

    if attribute is not None:
        if module.import_error is not None:
            return DataFrame([get_module_import_error_str(module)], columns=['error'])

        try:
            ml_handler = module.Handler(
                engine_storage=handlerStorage,
                model_storage=modelStorage
            )
            return ml_handler.describe(attribute)
        except NotImplementedError:
            return DataFrame()
        except Exception as e:
            return DataFrame(
                [f'{e.__class__.__name__}: {e}'],
                columns=['error']
            )
    else:
        model_info = ModelController.get_model_info(model_record)

        attrs_df = DataFrame()
        if module.import_error is not None:
            attrs_df = DataFrame(['import_error'], columns=['error'])
            model_error = model_info['ERROR'][0] or '-'
            model_info['ERROR'][0] = 'ML engine error:\n\n'
            model_info['ERROR'][0] += get_module_import_error_str(module)
            model_info['ERROR'][0] += '\nModel error:\n\n'
            model_info['ERROR'][0] += model_error
        else:
            try:
                ml_handler = module.Handler(
                    engine_storage=handlerStorage,
                    model_storage=modelStorage
                )
                attrs_df = ml_handler.describe(attribute)
            except NotImplementedError:
                pass
            except Exception as e:
                model_error = model_info['ERROR'][0] or '-'
                model_info['ERROR'][0] = 'ML engine error:\n\n'
                model_info['ERROR'][0] += f'{e.__class__.__name__}: {e}\n'
                model_info['ERROR'][0] += '\nModel error:\n\n'
                model_info['ERROR'][0] += model_error

        attributes = []
        if len(attrs_df) > 0 and len(attrs_df.columns) > 0:
            attributes = list(attrs_df[attrs_df.columns[0]])
            if len(attributes) == 1 and isinstance(attributes[0], list):
                # first cell already has a list
                attributes = attributes[0]

        model_info.insert(0, 'TABLES', [attributes])
        return model_info



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the PostgreSQL server.",
        "required": True,
        "label": "User",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the PostgreSQL server.",
        "required": True,
        "label": "Password",
        "secret": True
    },
    database={
        "type": ARG_TYPE.STR,
        "description": "The database name to use when connecting with the PostgreSQL server.",
        "required": True,
        "label": "Database",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the PostgreSQL server. NOTE: use '127.0.0.1' instead of 'localhost' to connect to local server.",
        "required": True,
        "label": "Host",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The TCP/IP port of the PostgreSQL server. Must be an integer.",
        "required": True,
        "label": "Port",
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "The schema in which objects are searched first.",
        "required": False,
        "label": "Schema",
    },
    sslmode={
        "type": ARG_TYPE.STR,
        "description": "sslmode that will be used for connection.",
        "required": False,
        "label": "sslmode",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1", port=5432, user="root", password="password", database="database"
)



from mindsdb.utilities.profiler.profiler import (
    start,
    stop,
    Context,
    profile,
    enable,
    disable,
    set_meta
)

__all__ = [
    'start',
    'stop',
    'Context',
    'profile',
    'enable',
    'disable',
    'set_meta'
]



from typing import List, Optional

import pinecone
import pandas as pd
import ast

from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class PineconeHandler(VectorStoreHandler):
    """This handler handles connection and execution of the Pinecone statements."""

    name = "pinecone"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.MAX_FETCH_LIMIT = 10000
        self._connection_data = kwargs.get("connection_data")
        self._client_config = {
            "api_key": self._connection_data.get("api_key"),
            "environment": self._connection_data.get("environment")
        }
        self._table_create_params = {
            "dimension": 8,
            "metric": "cosine",
            "pods": 1,
            "replicas": 1,
            "pod_type": 'p1',
        }
        for key in self._table_create_params:
            if key in self._connection_data:
                self._table_create_params[key] = self._connection_data[key]
        self.is_connected = False
        self.connect()

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def _get_index_handle(self, index_name):
        """Returns handler to index specified by `index_name`"""
        index = pinecone.Index(index_name)
        try:
            index.describe_index_stats()
        except Exception:
            index = None
        return index

    def _get_pinecone_operator(self, operator: FilterOperator) -> str:
        """Convert FilterOperator to an operator that pinecone's query language can undersand"""
        mapping = {
            FilterOperator.EQUAL: "$eq",
            FilterOperator.NOT_EQUAL: "$ne",
            FilterOperator.GREATER_THAN: "$gt",
            FilterOperator.GREATER_THAN_OR_EQUAL: "$gte",
            FilterOperator.LESS_THAN: "$lt",
            FilterOperator.LESS_THAN_OR_EQUAL: "$lte",
            FilterOperator.IN: "$in",
            FilterOperator.NOT_IN: "$nin",
        }
        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by Pinecone!")
        return mapping[operator]

    def _translate_metadata_condition(self, conditions: List[FilterCondition]) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by pinecone.
        E.g.,
        [
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.LESS_THAN,
                value="2020-01-01",
            ),
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.GREATER_THAN,
                value="2019-01-01",
            )
        ]
        -->
        {
            "$and": [
                {"created_at": {"$lt": "2020-01-01"}},
                {"created_at": {"$gt": "2019-01-01"}}
            ]
        }
        """
        # we ignore all non-metadata conditions
        if conditions is None:
            return None
        metadata_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value)
        ]
        if len(metadata_conditions) == 0:
            return None

        # we translate each metadata condition into a dict
        pinecone_conditions = []
        for condition in metadata_conditions:
            metadata_key = condition.column.split(".")[-1]
            pinecone_conditions.append(
                {
                    metadata_key: {
                        self._get_pinecone_operator(condition.op): condition.value
                    }
                }
            )

        # we combine all metadata conditions into a single dict
        metadata_condition = (
            {"$and": pinecone_conditions}
            if len(pinecone_conditions) > 1
            else pinecone_conditions[0]
        )
        return metadata_condition

    def _matches_to_dicts(self, matches: List):
        """Converts the custom pinecone response type to a list of python dict"""
        return [match.to_dict() for match in matches]

    def connect(self):
        """Connect to a pinecone database."""
        try:
            pinecone.init(api_key=self._client_config["api_key"], environment=self._client_config["environment"])
            pinecone.list_indexes()
            self.is_connected = True
        except Exception as e:
            logger.error(f"Error connecting to Pinecone client, {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the pinecone connection."""
        if self.is_connected is False:
            return
        pinecone.init(api_key="", environment="")
        self.is_connected = False

    def check_connection(self):
        """Check the connection to pinecone."""
        response_code = StatusResponse(False)
        try:
            pinecone.list_indexes()
            response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to pinecone , {e}!")
            response_code.error_message = str(e)
        return response_code

    def get_tables(self) -> HandlerResponse:
        """Get the list of indexes in the pinecone database."""
        indexes = pinecone.list_indexes()
        indexes_names = pd.DataFrame(
            columns=["index_name"],
            data=[index for index in indexes],
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=indexes_names)

    def create_table(self, table_name: str, if_not_exists=True):
        """Create an index with the given name in the Pinecone database."""
        pinecone.create_index(name=table_name, **self._table_create_params)

    def insert(self, table_name: str, data: pd.DataFrame, columns: List[str] = None):
        """Insert data into pinecone index passed in through `table_name` parameter."""
        upsert_batch_size = 99  # API reccomendation
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        data.rename(columns={
            TableField.ID.value: "id",
            TableField.EMBEDDINGS.value: "values",
            TableField.METADATA.value: "metadata"},
            inplace=True)
        data = data[["id", "values", "metadata"]]

        for chunk in (data[pos:pos + upsert_batch_size] for pos in range(0, len(data), upsert_batch_size)):
            chunk = chunk.to_dict(orient="records")
            index.upsert(vectors=chunk)

    def drop_table(self, table_name: str, if_exists=True):
        """Delete an index passed in through `table_name` from the pinecone ."""

        pinecone.delete_index(table_name)

    def delete(self, table_name: str, conditions: List[FilterCondition] = None):
        """Delete records in pinecone index `table_name` based on ids or based on metadata conditions."""
        filters = self._translate_metadata_condition(conditions)
        ids = [
            condition.value
            for condition in conditions
            if condition.column == TableField.ID.value
        ] or None
        if filters is None and ids is None:
            raise Exception("Delete query must have either id condition or metadata condition!")
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        if filters is None:
            index.delete(ids=ids)
        else:
            index.delete(filter=filters)

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ):
        """Run query on pinecone index named `table_name` and get results."""
        index = self._get_index_handle(table_name)
        if index is None:
            raise Exception(f"Error getting index '{table_name}', are you sure the name is correct?")

        query = {
            "include_values": True,
            "include_metadata": True
        }
        # check for metadata filter
        metadata_filters = self._translate_metadata_condition(conditions)
        # check for vector filter
        vector_filter = (
            None
            if conditions is None
            else [
                condition.value
                for condition in conditions
                if condition.column == TableField.SEARCH_VECTOR.value
            ]
        )
        if vector_filter:
            if len(vector_filter) > 1:
                raise Exception("You cannot have multiple search_vectors in query")

            query["vector"] = vector_filter[0]
            # For subqueries, the vector filter is a list of list of strings
            if isinstance(query["vector"], list) and isinstance(query["vector"][0], str):
                if len(query["vector"]) > 1:
                    raise Exception("You cannot have multiple search_vectors in query")

                try:
                    query["vector"] = ast.literal_eval(query["vector"][0])
                except Exception as e:
                    raise Exception(f"Cannot parse the search vector '{query['vector']}'into a list: {e}")

        # check for limit
        if limit is not None:
            query["top_k"] = limit
        else:
            query["top_k"] = self.MAX_FETCH_LIMIT
        if metadata_filters is not None:
            query["filter"] = metadata_filters
        # check for id filter
        id_filters = None
        if conditions is not None:
            id_filters = [
                condition.value
                for condition in conditions
                if condition.column == TableField.ID.value
            ] or None
        if id_filters:
            if len(id_filters) > 1:
                raise Exception("You cannot have multiple IDs in query")

            query["id"] = id_filters[0]
        # exec query
        try:
            result = index.query(**query)
        except Exception as e:
            raise Exception(f"Error running SELECT query on '{table_name}': {e}")

        # convert to dataframe
        df_columns = {
            "id": TableField.ID.value,
            "metadata": TableField.METADATA.value,
            "values": TableField.EMBEDDINGS.value,
        }
        results_df = pd.DataFrame.from_records(self._matches_to_dicts(result["matches"]))
        if bool(len(results_df.columns)):
            results_df.rename(columns=df_columns, inplace=True)
        else:
            results_df = pd.DataFrame(columns=list(df_columns.values()))
        results_df[TableField.CONTENT.value] = ""
        return results_df[columns]

    def get_columns(self, table_name: str) -> HandlerResponse:
        return super().get_columns(table_name)



from pydantic_settings import BaseSettings


class MindsEndpointHandlerConfig(BaseSettings):
    """
    Configuration for Minds Endpoint handler.

    Attributes
    ----------

    BASE_URL : str
        Base URL for the Minds Endpoint API.
    """

    BASE_URL: str = "https://llm.mdb.ai/"


minds_endpoint_handler_config = MindsEndpointHandlerConfig()



import unittest
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.handlers.datastax_handler.datastax_handler import (
    DatastaxHandler,
)


class DatastaxHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "user": "cassandra",
                "password": "",
                "secure_connect_bundle": "/home/Downloads/file.zip",
            }
        }
        cls.handler = DatastaxHandler("test_datastax_handler", **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_native_query_show_keyspaces(self):
        dbs = self.handler.native_query("DESC KEYSPACES;")
        assert dbs.type is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_3_describe_table(self):
        described = self.handler.get_columns("home_rentals")
        assert described.type is RESPONSE_TYPE.TABLE

    def test_4_select_query(self):
        query = "SELECT * FROM home_rentals WHERE 'id'='3712'"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE



from mindsdb.integrations.utilities.rag.rag_pipeline_builder import RAG
from mindsdb.integrations.utilities.rag.config_loader import load_rag_config

from mindsdb.interfaces.agents.constants import DEFAULT_EMBEDDINGS_MODEL_CLASS
from mindsdb.interfaces.skills.skill_tool import skill_tool
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.db import KnowledgeBase
from mindsdb.utilities import log
from langchain_core.tools import Tool
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import construct_model_from_args

logger = log.getLogger(__name__)


def build_retrieval_tool(tool: dict, pred_args: dict, skill: db.Skills):
    """
    Builds a retrieval tool i.e RAG

    Args:
        tool: Tool configuration dictionary
        pred_args: Predictor arguments dictionary
        skill: Skills database object

    Returns:
        Tool: Configured retrieval tool

    Raises:
        ValueError: If knowledge base is not found or configuration is invalid
    """
    # build RAG config
    tools_config = tool['config']
    tools_config.update(pred_args)

    kb_params = {}
    embeddings_model = None

    if 'source' in tool:
        kb_name = tool['source']
        executor = skill_tool.get_command_executor()
        kb = _get_knowledge_base(kb_name, skill.project_id, executor)

        if not kb:
            raise ValueError(f"Knowledge base not found: {kb_name}")

        kb_table = executor.session.kb_controller.get_table(kb.name, kb.project_id)
        kb_params = {
            'vector_store_config': {
                'kb_table': kb_table
            }
        }

        # Get embedding model from knowledge base table
        if kb_table._kb.embedding_model:
            # Extract embedding model args from knowledge base table
            embedding_args = kb_table._kb.embedding_model.learn_args.get('using', {})
            # Construct the embedding model directly
            embeddings_model = construct_model_from_args(embedding_args)
            logger.debug(f"Using knowledge base embedding model with args: {embedding_args}")
        else:
            embeddings_model = DEFAULT_EMBEDDINGS_MODEL_CLASS()
            logger.debug("Using default embedding model as knowledge base has no embedding model")
    elif 'embedding_model' not in tools_config:
        embeddings_model = DEFAULT_EMBEDDINGS_MODEL_CLASS()
        logger.debug("Using default embedding model as no knowledge base provided")

    # Load and validate config
    try:
        rag_config = load_rag_config(tools_config, kb_params, embeddings_model)
        # build retriever
        rag_pipeline = RAG(rag_config)
        logger.debug(f"RAG pipeline created with config: {rag_config}")

        def rag_wrapper(query: str) -> str:
            try:
                result = rag_pipeline(query)
                logger.debug(f"RAG pipeline result: {result}")
                return result['answer']
            except Exception as e:
                logger.error(f"Error in RAG pipeline: {str(e)}")
                return f"Error in retrieval: {str(e)}"

        # Create RAG tool
        return Tool(
            func=rag_wrapper,
            name=tool['name'],
            description=tool['description'],
            response_format='content',
            # Return directly by default since we already use an LLM against retrieved context to generate a response.
            return_direct=tools_config.get('return_direct', True)
        )

    except Exception as e:
        logger.error(f"Error building RAG pipeline: {str(e)}")
        raise ValueError(f"Failed to build RAG pipeline: {str(e)}")


def _get_knowledge_base(knowledge_base_name: str, project_id, executor) -> KnowledgeBase:
    """
    Get knowledge base by name and project ID

    Args:
        knowledge_base_name: Name of the knowledge base
        project_id: Project ID
        executor: Command executor instance

    Returns:
        KnowledgeBase: Knowledge base instance if found, None otherwise
    """
    kb = executor.session.kb_controller.get(knowledge_base_name, project_id)
    return kb



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': 'The AWS access key that identifies the user or IAM role.',
        'required': True,
        'label': 'AWS Access Key'
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': 'The AWS secret access key that identifies the user or IAM role.',
        'secret': True,
        'required': True,
        'label': 'AWS Secret Access Key'
    },
    region_name={
        'type': ARG_TYPE.STR,
        'description': 'The AWS region to connect to.',
        'required': True,
        'label': 'AWS Region'
    },
    aws_session_token={
        'type': ARG_TYPE.STR,
        'description': 'The AWS session token that identifies the user or IAM role. This becomes necessary when using temporary security credentials.',
        'secret': True,
        'required': False,
        'label': 'AWS Session Token'
    }
)

connection_args_example = OrderedDict(
    aws_access_key_id='PCAQ2LJDOSWLNSQKOCPW',
    aws_secret_access_key='U/VjewPlNopsDmmwItl34r2neyC6WhZpUiip57i',
    region_name='us-east-1'
)



"""db files

Revision ID: 27c5aca9e47e
Revises: 47f97b83cee4
Create Date: 2022-02-09 10:43:29.854671

"""
import json
import datetime

from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db
from sqlalchemy.sql import text


# revision identifiers, used by Alembic.
revision = '27c5aca9e47e'
down_revision = '47f97b83cee4'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('ai_table')

    conn = op.get_bind()

    # views was created with unnamed fk. Therefore need recreate it
    op.create_table(
        'view_tmp',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('query', sa.String(), nullable=False),
        sa.Column('integration_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['integration_id'], ['integration.id'], name='fk_integration_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_view_name_company_id')
    )
    conn.execute(text("""
        insert into view_tmp (id, name, company_id, query, integration_id)
        select id, name, company_id, query, datasource_id from view;
    """))
    op.drop_table('view')
    op.rename_table('view_tmp', 'view')

    op.create_table(
        'analysis',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('analysis', mindsdb.interfaces.storage.db.Json(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    with op.batch_alter_table('datasource', schema=None) as batch_op:
        batch_op.add_column(sa.Column('analysis_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key('fk_ds_analysis_id', 'analysis', ['analysis_id'], ['id'])
        batch_op.add_column(sa.Column('ds_class', sa.String(), nullable=True))

    session = sa.orm.Session(bind=conn)
    dsatasources = conn.execute(sa.text('select id, analysis from datasource')).fetchall()
    for row in dsatasources:
        if row['analysis'] is not None:
            # NOTE 'returning' is relatively new in sqlite, so better will be use select after insert.
            conn.execute(
                text("""
                    insert into analysis (analysis) select analysis from datasource where id = :id;
                """), {
                    'id': row['id']
                }
            )
            analysis_id = conn.execute(text("""
                select id from analysis order by id desc limit 1;
            """)).fetchall()
            conn.execute(
                text("""
                    update datasource set analysis_id = :analysis_id where id = :id
                """), {
                    'analysis_id': analysis_id[0][0],
                    'id': row['id']
                }
            )

    with op.batch_alter_table('datasource', schema=None) as batch_op:
        batch_op.drop_column('analysis')

    op.create_table(
        'file',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('source_file_path', sa.String(), nullable=False),
        sa.Column('file_path', sa.String(), nullable=False),
        sa.Column('row_count', sa.Integer(), nullable=False),
        sa.Column('columns', mindsdb.interfaces.storage.db.Json(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True, server_default=sa.func.current_timestamp()),
        sa.Column('updated_at', sa.DateTime(), nullable=True, server_default=sa.func.current_timestamp(), server_onupdate=sa.func.current_timestamp()),
        sa.Column('analysis_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['analysis_id'], ['analysis.id'], name='fk_analysis_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_file_name_company_id')
    )

    # delete ds where data is none
    dsatasources = conn.execute(text('select * from datasource')).fetchall()
    for ds in dsatasources:
        if ds['data'] is None:
            conn.execute(text('delete from datasource where id = :id'), {'id': ds['id']})
            continue
        ds_data = json.loads(ds['data'])
        creation_info = json.loads(ds['creation_info'])
        datasource_name = ds_data.get('source_type')
        if datasource_name == 'file':
            created_at = None
            if isinstance(ds['created_at'], str):
                created_at = datetime.datetime.fromisoformat(ds['created_at'])
            elif isinstance(ds['created_at'], [float, int]):
                created_at = datetime.fromtimestamp(ds['created_at'])

            updated_at = None
            if isinstance(ds['updated_at'], str):
                updated_at = datetime.datetime.fromisoformat(ds['updated_at'])
            elif isinstance(ds['updated_at'], [float, int]):
                updated_at = datetime.fromtimestamp(ds['updated_at'])

            file = mindsdb.interfaces.storage.db.File(
                name=ds['name'],
                company_id=ds['company_id'],
                source_file_path=ds_data['source'],
                file_path=creation_info['args'][0],
                row_count=ds_data['row_count'],
                columns=ds_data['columns'],
                created_at=created_at,
                updated_at=updated_at,
                analysis_id=ds['analysis_id']
            )
            session.add(file)
            session.flush()
            # ds_data['file_id'] = file.id
            ds_data['source'] = {
                'mindsdb_file_name': ds['name']
                # 'source': ds_data['source']
            }
            conn.execute(
                text("""
                    update datasource set data = :ds_data where id = :id;
                """), {
                    'id': ds['id'],
                    'ds_data': json.dumps(ds_data)
                }
            )

        conn.execute(
            text("""
                update datasource
                set integration_id = (select id from integration where name = :datasource_name and company_id = :company_id),
                    ds_class = :ds_class
                where id = :id
            """), {
                'datasource_name': datasource_name,
                'company_id': ds['company_id'],
                'ds_class': creation_info['class'],
                'id': ds['id']
            }
        )

    session.commit()

    op.rename_table('datasource', 'dataset')

    with op.batch_alter_table('dataset', schema=None) as batch_op:
        batch_op.create_foreign_key('fk_integration_id', 'integration', ['integration_id'], ['id'])

    # NOTE two different 'batch' is necessary, in other way FK is not creating
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column('datasource_id', new_column_name='dataset_id')
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_foreign_key('fk_predictor_dataset_id', 'dataset', ['dataset_id'], ['id'])
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_unique_constraint('unique_predictor_name_company_id', ['name', 'company_id'])

    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.create_unique_constraint('unique_integration_name_company_id', ['name', 'company_id'])

    with op.batch_alter_table('dataset', schema=None) as batch_op:
        batch_op.create_unique_constraint('unique_dataset_name_company_id', ['name', 'company_id'])


def downgrade():
    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.drop_constraint('unique_integration_name_company_id', type_='unique')

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('unique_predictor_name_company_id', type_='unique')

    with op.batch_alter_table('dataset', schema=None) as batch_op:
        batch_op.drop_constraint('unique_dataset_name_company_id', type_='unique')

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('fk_predictor_dataset_id', type_='foreignkey')
        batch_op.alter_column('dataset_id', new_column_name='datasource_id')

    with op.batch_alter_table('dataset', schema=None) as batch_op:
        batch_op.drop_constraint('fk_integration_id', type_='foreignkey')
        batch_op.add_column(sa.Column('analysis', sa.VARCHAR(), nullable=True))
        batch_op.drop_constraint('fk_ds_analysis_id', type_='foreignkey')
        batch_op.drop_column('ds_class')

    op.rename_table('dataset', 'datasource')

    op.drop_table('file')

    conn = op.get_bind()
    conn.execute(text("""
        update datasource set analysis = (select analysis from analysis where id = analysis_id)
    """))

    with op.batch_alter_table('datasource', schema=None) as batch_op:
        batch_op.drop_column('analysis_id')

    op.drop_table('analysis')

    op.create_table(
        'ai_table',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('integration_name', sa.String(), nullable=False),
        sa.Column('integration_query', sa.String(), nullable=False),
        sa.Column('query_fields', mindsdb.interfaces.storage.db.Json(), nullable=False),
        sa.Column('predictor_name', sa.String(), nullable=False),
        sa.Column('predictor_columns', mindsdb.interfaces.storage.db.Json(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'view_tmp',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('query', sa.String(), nullable=False),
        sa.Column('datasource_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['datasource_id'], ['integration.id'], name='fk_datasource_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_name_company_id')
    )
    conn.execute(text("""
        insert into view_tmp (id, name, company_id, query, datasource_id)
        select id, name, company_id, query, integration_id from view;
    """))
    op.drop_table('view')
    op.rename_table('view_tmp', 'view')



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    server_name={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the database'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect.'
    },
    database_name={
        'type': ARG_TYPE.STR,
        'description': '''
            The database name to use when connecting.
        '''
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate.',
        'secret': True
    },
    
)

connection_args_example = OrderedDict(
    server_name = 'localhost',
    port =  9001,
    database_name = 'xdb',
    username = 'SA',
    password = 'password'
)



from typing import Text, List
from pydantic_settings import BaseSettings


class AnyscaleHandlerConfig(BaseSettings):
    """
    Configuration for the Anyscale handler.

    Attributes
    ----------

    ANYSCALE_API_BASE : Text
        Base URL for the Anyscale API.

    MIN_FT_VAL_LEN : int
        Minimum number of validation chats required for fine-tuning.

    MIN_FT_DATASET_LEN : int
        Minimum number of training and validation chats required for fine-tuning.

    DEFAULT_MODEL : Text
        Default model to use for models.

    DEFAULT_MODE : Text
        Default mode to use for models. Can be 'default', 'conversational', or 'conversational-full'.

    SUPPORTED_MODES : List
        List of supported modes. Can be 'default', 'conversational', or 'conversational-full'.

    RATE_LIMIT : int
        Number of requests per minute.

    MAX_BATCH_SIZE : int
        Maximum batch size for requests.

    DEFAULT_MAX_TOKENS : int
        Default maximum tokens for requests.
    """

    ANYSCALE_API_BASE: Text = 'https://api.endpoints.anyscale.com/v1'
    MIN_FT_VAL_LEN: int = 20
    MIN_FT_DATASET_LEN: int = MIN_FT_VAL_LEN * 2
    DEFAULT_MODEL: Text = 'meta-llama/Llama-2-7b-chat-hf'
    DEFAULT_MODE: Text = 'default'
    SUPPORTED_MODES: List = ['default', 'conversational', 'conversational-full']
    RATE_LIMIT: int = 25
    MAX_BATCH_SIZE: int = 20
    DEFAULT_MAX_TOKENS: int = 100


anyscale_handler_config = AnyscaleHandlerConfig()



import os

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider, Span
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from mindsdb.utilities import log
logger = log.getLogger(__name__)


# Check OpenTelemetry exporter type
OTEL_EXPORTER_TYPE = os.getenv("OTEL_EXPORTER_TYPE", "console")  # console or otlp

# Define OTLP endpoint. If not set, the default OTLP endpoint will be used
OTEL_OTLP_ENDPOINT = os.getenv("OTEL_OTLP_ENDPOINT", "http://localhost:4317")

# Define service name
OTEL_SERVICE_NAME = os.getenv("OTEL_SERVICE_NAME", "mindsdb")

# The name of the environment we"re on, by default local for development, this is set differently per-env in our Helm chart values files
OTEL_SERVICE_ENVIRONMENT = os.getenv("OTEL_SERVICE_ENVIRONMENT", "local").lower()

# Define service release
OTEL_SERVICE_RELEASE = os.getenv("OTEL_SERVICE_RELEASE", "local").lower()

# By default we have Open Telemetry SDK enabled on all envs, except for local which is disabled by default
#   If you want to enable Open Telemetry on local for some reason please set OTEL_SDK_FORCE_RUN to true
OTEL_SDK_DISABLED = os.getenv("OTEL_SDK_DISABLED", "false").lower() == "true" or os.getenv("OTEL_SERVICE_ENVIRONMENT", "local").lower() == "local"
OTEL_SDK_FORCE_RUN = os.getenv("OTEL_SDK_FORCE_RUN", "false").lower() == "true"

# Custom span processor to add global tags to spans


class GlobalTaggingSpanProcessor(BatchSpanProcessor):
    def on_start(self, span: Span, parent_context):
        # Add environment and release to every span
        span.set_attribute("environment", OTEL_SERVICE_ENVIRONMENT)
        span.set_attribute("release", OTEL_SERVICE_RELEASE)
        super().on_start(span, parent_context)


if not OTEL_SDK_DISABLED or OTEL_SDK_FORCE_RUN:
    logger.info("OpenTelemetry enabled")
    logger.info(f"OpenTelemetry exporter type: {OTEL_EXPORTER_TYPE}")
    logger.info(f"OpenTelemetry service name: {OTEL_SERVICE_NAME}")
    logger.info(f"OpenTelemetry service environment: {OTEL_SERVICE_ENVIRONMENT}")
    logger.info(f"OpenTelemetry service release: {OTEL_SERVICE_RELEASE}")

    # Define OpenTelemetry resources (e.g., service name)
    resource = Resource(attributes={"service.name": OTEL_SERVICE_NAME})

    # Set the tracer provider with the custom resource
    trace.set_tracer_provider(TracerProvider(resource=resource))

    # Configure the appropriate exporter based on the environment variable
    if OTEL_EXPORTER_TYPE == "otlp":
        logger.info("OpenTelemetry is using OTLP exporter")

        exporter = OTLPSpanExporter(
            endpoint=OTEL_OTLP_ENDPOINT,  # Default OTLP endpoint
            insecure=True  # Disable TLS for local testing
        )

    else:
        logger.info("OpenTelemetry is using Console exporter")

        exporter = ConsoleSpanExporter()

    # Replace the default span processor with the custom one
    trace.get_tracer_provider().add_span_processor(GlobalTaggingSpanProcessor(exporter))
